package net.sf.gluebooster.demos.pojo.planning;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

import javax.naming.Context;
import javax.naming.Name;
import javax.naming.spi.ObjectFactory;

import net.sf.gluebooster.demos.pojo.refactor.BoostedNodeGraph;
import net.sf.gluebooster.demos.pojo.refactor.BoostedObjectFactory;
import net.sf.gluebooster.demos.pojo.refactor.CollectionSelector;
import net.sf.gluebooster.demos.pojo.refactor.Condition;
import net.sf.gluebooster.demos.pojo.refactor.ObjectAttributes;
import net.sf.gluebooster.demos.pojo.refactor.Operation;
import net.sf.gluebooster.demos.pojo.refactor.OperationSorter;
import net.sf.gluebooster.java.booster.basic.container.BoostedNode;

import org.apache.commons.collections15.Transformer;
import org.apache.commons.lang3.tuple.Pair;

/**
 * Meta planning algorithms.
 * 
 * The ObjectFactory has the signature Planning getObjectInstance (Planning, ...)
 * 
 * @author CBauer
 * @defaultParamText planning the current state of the planning with all necessary data
 *
 * 
 */
public class MetaPlanner<Conditio extends Condition<Conditio, NameOfObjects>, NameOfObjects>
		extends
		BoostedObjectFactory<Planning<Conditio, NameOfObjects>, Planning<Conditio, NameOfObjects>>
		implements ObjectFactory {

	/**
	 * The variant that does nothing
	 */
	private static final int VARIANT_IDLE = 0;

	/**
	 * One (the first) planning algorithm
	 */
	private static final int VARIANT_1 = 1;

	/**
	 * A variant that just expands not but does not solve conflicts.
	 */
	private static final int VARIANT_JUST_EXPAND = 2;

	/**
	 * The description of the algorithm of the planning.
	 */
	private int variant = VARIANT_IDLE;

	/**
	 * Selects one of the expanders that are selected with World.getExpander
	 * Signature is ObjectFactory getObjectInstance(Collection<ObjectFactory>,
	 * ...)
	 */
	private Transformer<Collection<Transformer<BoostedNode, BoostedNodeGraph>>, Transformer<BoostedNode, BoostedNodeGraph>> planNodeExpanderSelector = CollectionSelector
			.createSelectFirstElementSelector();

	public MetaPlanner(int variant) {
		this.variant = variant;
	}

	/**
	 * Create a planner that does nothing.
	 * 
	 * @return the created planner
	 */
	public static /*<Conditio extends Condition<Conditio, NameOfObjects>, NameOfObjects>*/ Transformer/*<Planning<Conditio, NameOfObjects>, Planning<Conditio, NameOfObjects>>*/ createIdlePlanner() {
		return new MetaPlanner/*<Conditio, NameOfObjects>*/(VARIANT_IDLE);
	}

	/**
	 * Create a planner with a default planning algorithm
	 * 
	 * @return the created planner
	 */
	public static /*<Conditio extends Condition<Conditio, NameOfObjects>, NameOfObjects>*/ Transformer/*<Planning<Conditio, NameOfObjects>, Planning<Conditio, NameOfObjects>>*/ createPlanner1() {
		return new MetaPlanner/*<Conditio, NameOfObjects>*/(VARIANT_1);
	}

	/**
	 * Create a planner that just expands the nodes, but does no conflict solving, etc.
	 * 
	 * @return the created planner
	 */
	public static /*<Conditio extends Condition<Conditio, NameOfObjects>, NameOfObjects>*/ Transformer/*<Planning<Conditio, NameOfObjects>, Planning<Conditio, NameOfObjects>>*/ createPlannerThatJustExpands() {
		return new MetaPlanner/*<Conditio, NameOfObjects>*/(VARIANT_JUST_EXPAND);
	}

	/**
	 * Does the planning
	 * 
	 * @param planningInstance
	 *            instance of Planning
	 * @return Planning
	 */
	@Override
	public Object getObjectInstance(Object planningInstance, Name name,
			Context nameCtx, Hashtable<?, ?> environment) throws Exception {

		Planning<Conditio, NameOfObjects> planning = (Planning<Conditio, NameOfObjects>) planningInstance;
		switch (variant) {
		case VARIANT_IDLE:
			planIdle(planning);
			break;
		case VARIANT_1:
			plan1(planning);
			break;
		case VARIANT_JUST_EXPAND:
			planJustExpand(planning);
			break;
		default:
			throw new IllegalStateException("variant " + variant
					+ " not (yet) supported");
		}
		return planningInstance;
	}


	/**
	 * Does nothing
	 * 
	 * @param configuration
	 *            the planning
	 */
	private void planIdle(Planning<Conditio, NameOfObjects> configuration)
			throws Exception {
		configuration.info("idle planner, no planning is done");
	}

	/**
	 * Algorithm 1
	 * 
	 * @param planning
	 */
	private void plan1(Planning<Conditio, NameOfObjects> planning)
			throws Exception {
		// planning.debug("planner 1");
		
		int i = 0;
		int maxSteps = planning.getEnv().getMaxSteps();

		while (i < maxSteps && !planning.getPlan().isSolved()) {
			i++;


			planning.info("planner ", variant, ": step ", i, " of at most ",
					maxSteps);

			// TODO use a time dependent combineNodes instead of i
			if (!nextStep1(planning, i < 30)) {
				simplify1(planning);
			}

			if (planning.getPlan().isSolved())
				planning.getPlan().removeSuperfluousOperators();

			if (planning.getPlan().getVersion() == null) {
				planning.getPlan().setVersion(1);
			} else {
				planning.getPlan().setVersion(
						planning.getPlan().getVersion().intValue() + 1);
			}

		}
		

	}

	/**
	 * Algorithm 1
	 * 
	 * @param plan
	 * @param env
	 * @throws Exception
	 */
	private void planJustExpand(Planning<Conditio, NameOfObjects> planning)
			throws Exception {

		int i = 0;
		int maxSteps = planning.getEnv().getMaxSteps();
		planning.getPlan().setVersion(1);

		while (i < maxSteps && !planning.getPlan().isSolved()) {
			i++;

			planning.info("planner ", variant, ": step ", i, " of at most ",
					maxSteps);

			nextStepJustExpandOneNode(planning);


			planning.getPlan().setVersion(
					planning.getPlan().getVersion().intValue() + 1);

		}

	}

	/**
	 * Do one step (of planning algorithm 1)
	 * 
	 * @param combineNodes
	 *            the nodes to combine
	 * @return true, if the plan has been modified
	 */
	public boolean nextStep1(Planning<Conditio, NameOfObjects> planning,
			boolean combineNodes) throws Exception {


		boolean result = instantiateOneRandomVirtualObject(planning);

		result |= solveSomeUnsolvedNodeWithMaximumAbstraction(planning, 10);


		if (combineNodes) {
			result |= combineNodes(planning);
			while (combineNodes(planning)) {
				// just repeat the combining process as long as possible
			}
		}

		result |= solveEarlyConflictSeries(planning);
		while (solveEarlyConflictSeries(planning)) {
			// just repeat the solving process as long as possible
		}

		if (!result) {
			result = solveGoalConflict(planning);
		}

		if (!result) {
			// plan->combine_above = 0;
			result = combineNodes(planning);
			// plan->combine_above = 1;
		}

		return result;

	}

	/**
	 * Do one step (of planning algorithm 1)
	 * 
	 * @param plan
	 * @return true, if the plan has been modified
	 */
	public boolean nextStepJustExpandOneNode(
			Planning<Conditio, NameOfObjects> planning)
			throws Exception {

		boolean result = planning.removeSuperfluousOperators();

		if (!result) {
			result = instantiateOneRandomVirtualObject(planning);
		}

		if (!result) {
			result = solveSomeUnsolvedNodeWithMaximumAbstraction(planning, 1);
		}

		return result;
	}

	/**
	 * Tries to solve the nodes with the highest abstraction level.
	 * 
	 * If the highest abstraction level is not solvable continue with the next lower level. Try to solve all nodes with the same abstraction level. If any node
	 * is solved, do not continue with the next lower abstraction level, but stop
	 * 
	 * @param maxNumberOfNodes
	 *            the maximum number of nodes to solve/expand
	 * @return true, if the plan has been modified
	 * @throws Exception
	 */
	public boolean solveSomeUnsolvedNodeWithMaximumAbstraction(
			Planning<Conditio, NameOfObjects> planning, int maxNumberOfNodes)
			throws Exception {

		int abstractionLevel = 0;
		boolean solved = false;

		Collection<BoostedNode> unsolvedNodes = planning.getPlan()
				.getUnsolvedNodes(
						OperationSorter.createSortByOperatorAbstractionDesc());
		int counter = 0;
		for (BoostedNode unsolvedNode : unsolvedNodes) {
			int newAbstractionLevel = Operation.getOperation(unsolvedNode)
					.getOperator()
					.getAbstractionLevel();
			if (solved) {
				if (newAbstractionLevel < abstractionLevel) {
					break; // stop solving
				}
			} else {
				abstractionLevel = newAbstractionLevel;
			}
			if (counter < maxNumberOfNodes) {
				boolean tempResult = expandUnsolvedNode(planning, unsolvedNode);
				solved |= tempResult;
				if (tempResult) {
					counter++;
				}
			} else {
				break;
			}
		}

		return solved;



	}

	/**
	 * Expands an unsolved node to a graph with of (easier solvable ) nodes. The plan is modified
	 * 
	 * @param planning
	 *            contains the plan
	 * @param unsolvedNode
	 *            the node to expland
	 * @return true if something has been changed
	 * @throws Exception
	 */
	public boolean expandUnsolvedNode(
			Planning<Conditio, NameOfObjects> planning,
			BoostedNode unsolvedNode) throws Exception {

		// Refactor.checkPredecessorSuccessor(unsolvedNode);

		Collection<Transformer<BoostedNode, BoostedNodeGraph>> solvers = planning
				.getPlan().getWorld()
				.getExpander(unsolvedNode);
		if (solvers.isEmpty()) {
			return false;
		} else {
			boolean result = planning.transformNode(unsolvedNode,
					planNodeExpanderSelector.transform(solvers));

			// planning.getPlan().simpleDisplay();
			return result;
		}

	}

	/**
	 * If nodes do the same operation in parallel branches, it may be possible
	 * merge the two operations into one by changing the plan structure. One
	 * operation is removed, its predecessors become predecessors of the other
	 * operation; its successors become successors of the other operation.
	 * 
	 * @return true, if the plan has been modified
	 */
	public boolean combineNodes(Planning<Conditio, NameOfObjects> planning) {

		// Algorithm
		// first get all nodes with equal operators
		// try to merge two of them.

		// first get all nodes with equal operators

		Set<Set<BoostedNode>> setsOfnodesWithEqualOperators = planning
				.getPlan()
				.getNodesWithEqualOperations();

		// get pairs of the nodes ant try to combine them until something was
		// sucessful or you run out of nodes
		for (Set<BoostedNode> nodesWithEqualOperators : setsOfnodesWithEqualOperators) {
			ArrayList<BoostedNode> otherNodes = new ArrayList<BoostedNode>(
					nodesWithEqualOperators);
			for (BoostedNode node : nodesWithEqualOperators) {
				otherNodes.remove(node); // so that there is no duplicate
				for (BoostedNode otherNode : otherNodes) {
					if (planning.getPlan().combineNodesWithEqualOperators(node,
							otherNode)) {
						return true;
					}
				}
			}
		}

		return false;
	}

	/**
	 * Solves the possible conflicts in the order of their graph position
	 * 
	 * @return true, if the plan has been modified
	 */
	public boolean solveEarlyConflictSeries(
			Planning<Conditio, NameOfObjects> planning) throws Exception {

		List<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>> possibleConflicts = planning
				.getPlan().orderPossibleConflictsByGraphPosition();

		for (ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>> possibleConflict : possibleConflicts) {
			if (planning.getPlan().solvePossibleConflictByReordering(
					possibleConflict)) {
				return true;
			}
		}

		return false;


	}

	/**
	 * Solve a goal conflict.
	 * 
	 * @return true, if the plan has been modified
	 */
	public boolean solveGoalConflict(Planning<Conditio, NameOfObjects> planning) {
		planning.error("solveGoalConflict is not yet implemented");
		// TODO do something
		// updateGlobalConditions();
		return false;
	}

	/**
	 * Instantiates a temporary object with a random instance.
	 * 
	 * @return true, if the plan has been modified
	 */
	public boolean instantiateOneRandomVirtualObject(
			Planning<Conditio, NameOfObjects> planning)
			throws Exception {

		Plan<Conditio, NameOfObjects> plan = planning.getPlan();

		Set<BoostedNode> nodes = plan.getNodesWithVirtualProperties();
		if (nodes.isEmpty()) {
			return false;
		} else {
			// take the first object and instantiate it with a random instance
			BoostedNode node = nodes.iterator().next();
			Set<ObjectFactory> initializers = plan.getWorld()
					.getVirtualObjectInitializer(node);
			Pair<WorldObject<NameOfObjects, ?>, WorldObject<NameOfObjects, ?>> virtualAndConcreteObject = (Pair<WorldObject<NameOfObjects, ?>, WorldObject<NameOfObjects, ?>>) initializers
					.iterator().next()
					.getObjectInstance(node, null, null,
							Planning.createHashtableWithPlan(plan));

			if (virtualAndConcreteObject != null) {
				planning.instantiateVirtualObject(node,
						virtualAndConcreteObject.getLeft(),
						virtualAndConcreteObject.getRight());
				return true;
			}
			else
				return false;
		}


	}

	/**
	 * Simplify a plan by replacing subtrees with new nodes that are to be
	 * solved. Simplifies the plan according to planning algorithm 1.
	 * 
	 * @param plan
	 */
	public void simplify1(Planning<Conditio, NameOfObjects> planning) {
		planning.error("simplify1 is not yet implemented");
		// TODO implement plan->undo();
	}


}
