package net.sf.gluebooster.demos.pojo.planning;

import java.awt.Color;
import java.awt.Paint;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.naming.spi.ObjectFactory;
import javax.xml.transform.ErrorListener;
import javax.xml.transform.TransformerException;

import net.sf.gluebooster.demos.pojo.planning.blockworld.Table;
import net.sf.gluebooster.demos.pojo.refactor.BoostedComparator;
import net.sf.gluebooster.demos.pojo.refactor.BoostedNodeGraph;
import net.sf.gluebooster.demos.pojo.refactor.ComparatorBoostUtils;
import net.sf.gluebooster.demos.pojo.refactor.Condition;
import net.sf.gluebooster.demos.pojo.refactor.ConditionBasics;
import net.sf.gluebooster.demos.pojo.refactor.GraphDisplayConfiguration;
import net.sf.gluebooster.demos.pojo.refactor.ObjectAttributes;
import net.sf.gluebooster.demos.pojo.refactor.Operation;
import net.sf.gluebooster.demos.pojo.refactor.Operator;
import net.sf.gluebooster.demos.pojo.refactor.Refactor;
import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.basic.container.BoostedNodeUtils;
import net.sf.gluebooster.java.booster.basic.container.SimpleBoostedNode;
import net.sf.gluebooster.java.booster.basic.math.TransformationGraph;
import net.sf.gluebooster.java.booster.essentials.meta.HasName;
import net.sf.gluebooster.java.booster.essentials.objects.BoostedObject;
import net.sf.gluebooster.java.booster.essentials.utils.Check;
import net.sf.gluebooster.java.booster.essentials.utils.ContainerBoostUtils;

import org.apache.commons.collections15.Transformer;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

/**
 * The data container that contains the details of a plan.
 * 
 * @author CBauer
 * 
 * @defaultParamText world the world the planning takes place in
 * @defaultParamText updatedNodes nodes that have been updated
 * @defaultParamText removedNodes nodes that have been removed
 * @defaultParamText node1 the first node
 * @defaultParamText node2 the other node
 * @defaultParamText additionalPrecondition the precondition to be added
 * 
 *
 */
public class Plan<Conditio extends Condition<Conditio, NameOfObjects>, NameOfObjects>
		extends
 BoostedObject implements ErrorListener {

	/**
	 * The world this plan is meant for.
	 */
	private World<Conditio, NameOfObjects> world;

	/**
	 * The description of the objects of the world.
	 */
	private Conditio worldInstanceDescription;

	/**
	 * The version of this plan.
	 */
	private Integer version;

	/**
	 * A node of the graph of the plan. The value attribute contains a Operation
	 */
	private BoostedNodeGraph graph;

	/**
	 * Pairs of nodes that are in a potential conflict. The name is the name of
	 * the conflict, the type is the type of the conflict.
	 */
	private Set<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>> possibleConflicts;

	/**
	 * A flag to indicate that a warning occured.
	 */
	private boolean warningHappened = false;

	/**
	 * A flag to indicate that an error occured.
	 */
	private boolean errorHappened = false;

	/**
	 * A flag to indicate that a fatal error occured.
	 */
	private boolean fatalErrorHappened = false;

	public Plan() {
	}

	/**
	 * Creates a plan with one to-solve node.
	 * 
	 * @param globalPrecondition
	 *            the initial situation in the world
	 * @param postcondition
	 *            the target situation in the world
	 */
	public Plan(World<Conditio, NameOfObjects> world,
			Conditio globalPrecondition,
			Conditio postcondition) throws Exception {
		worldInstanceDescription = globalPrecondition;
		initialize(world, globalPrecondition, postcondition);

	}

	/**
	 * Orders the possibleConflicts so that nodes that the plan structure is
	 * taken into consideration. Example: if node A is predecessor of node B
	 * then conflicts of node a appear in the list before the conflicts of node
	 * B.
	 * 
	 * @return an ordering of the possibleConflicts
	 */
	public List<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>> orderPossibleConflictsByGraphPosition() {
		ArrayList<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>> result = new ArrayList<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>>(
				possibleConflicts);

		BoostedComparator<BoostedNode> nodeComparator = ComparatorBoostUtils
				.createComparatorByListPosition(graph.getAllVerticesOrdered());

		Transformer<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>, Pair<BoostedNode, BoostedNode>> pairExtractor = new Transformer<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>, Pair<BoostedNode, BoostedNode>>() {

			@Override
			public Pair<BoostedNode, BoostedNode> transform(
					ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>> input) {

				return input.getValue();
			}
		};

		@SuppressWarnings("unchecked")
		Comparator pairComparator = ComparatorBoostUtils.createPairComparator(
				(Transformer) pairExtractor, nodeComparator, true);

		Refactor.sortPartialOrderedList(result, pairComparator);

		return result;

		// Comparator comparator = new
		// Comparator<ObjectAttributesDuplicate<String, String,
		// Pair<BoostedNode, BoostedNode>>>() {
		// /**
		// *
		// * There is only a partial ordering of the conflict pairs according
		// * to their position.
		// *
		// */
		// @Override
		// public int compare(
		// ObjectAttributesDuplicate<String, String, Pair<BoostedNode,
		// BoostedNode>> conflict1,
		// ObjectAttributesDuplicate<String, String, Pair<BoostedNode,
		// BoostedNode>> conflict2) {
		// try {
		// BoostedNode node1a = conflict1.getValue().getLeft();
		//
		// BoostedNode node1b = conflict1.getValue().getRight();
		//
		// BoostedNode node2a = conflict2.getValue().getLeft();
		//
		// BoostedNode node2b = conflict2.getValue().getRight();
		//
		// boolean node1aIsPredecessorOfNode2a = nodeComparator
		// .compare(node1a, node2a) < 0;
		// boolean node1aIsPredecessorOfNode2b = nodeComparator
		// .compare(node1a, node2b) < 0;
		// boolean node1bIsPredecessorOfNode2a = nodeComparator
		// .compare(node1b, node2a) < 0;
		// boolean node1bIsPredecessorOfNode2b = nodeComparator
		// .compare(node1b, node2b) < 0;
		// boolean node1aIsSuccessorOfNode2a = nodeComparator.compare(
		// node1a, node2a) > 0;
		// boolean node1aIsSuccessorOfNode2b = nodeComparator.compare(
		// node1a, node2b) > 0;
		// boolean node1bIsSuccessorOfNode2a = nodeComparator.compare(
		// node1b, node2a) > 0;
		// boolean node1bIsSuccessorOfNode2b = nodeComparator.compare(
		// node1b, node2b) > 0;
		//
		// if (node1aIsSuccessorOfNode2a
		// && node1aIsPredecessorOfNode2b)
		// return 0;
		// if (node1aIsSuccessorOfNode2b
		// && node1aIsPredecessorOfNode2a)
		// return 0;
		// if (node1bIsSuccessorOfNode2a
		// && node1bIsPredecessorOfNode2b)
		// return 0;
		// if (node1bIsSuccessorOfNode2b
		// && node1bIsPredecessorOfNode2a)
		// return 0;
		// if (node1aIsPredecessorOfNode2a
		// && node1bIsSuccessorOfNode2a)
		// return 0;
		// if (node1aIsPredecessorOfNode2b
		// && node1bIsSuccessorOfNode2b)
		// return 0;
		// if (node1bIsPredecessorOfNode2a
		// && node1aIsSuccessorOfNode2a)
		// return 0;
		// if (node1bIsPredecessorOfNode2b
		// && node1aIsSuccessorOfNode2b)
		// return 0;
		//
		// if (node1aIsPredecessorOfNode2a) {
		// if (node1bIsSuccessorOfNode2b) {
		// return 0;
		// } else {
		// return -1;
		// }
		// }
		//
		// if (node1aIsPredecessorOfNode2b) {
		// if (node1bIsSuccessorOfNode2a) {
		// return 0;
		// } else {
		// return -1;
		// }
		// }
		//
		// if (node1bIsPredecessorOfNode2a) {
		// if (node1aIsSuccessorOfNode2b) {
		// return 0;
		// } else {
		// return -1;
		// }
		// }
		//
		// if (node1bIsPredecessorOfNode2b) {
		// if (node1aIsSuccessorOfNode2a) {
		// return 0;
		// } else {
		// return -1;
		// }
		// }
		//
		// if (node1aIsSuccessorOfNode2a || node1aIsSuccessorOfNode2b
		// || node1bIsSuccessorOfNode2a
		// || node1bIsSuccessorOfNode2b) {
		// return 1;
		// }
		//
		// return 0;
		// } catch (Exception ex) {
		// throw ThrowableBoostUtils.toRuntimeException(ex);
		// }
		// }
		//
		// private int compareOld(
		// ObjectAttributesDuplicate<String, String, Pair<BoostedNode,
		// BoostedNode>> conflict1,
		// ObjectAttributesDuplicate<String, String, Pair<BoostedNode,
		// BoostedNode>> conflict2) {
		// try {
		// int hash1 = conflict1.hashCode();
		// BoostedNode node1a = conflict1.getValue().getLeft();
		// List<BoostedNode> node1aPredecessor = BoostedNodeUtils
		// .getPredecessors(node1a);
		// List<BoostedNode> node1aSuccessor = BoostedNodeUtils
		// .getSuccessors(node1a);
		// // getLog().info(Refactor.getStructureString(node1a));
		//
		// int hash2 = conflict2.hashCode();
		// BoostedNode node1b = conflict1.getValue().getRight();
		// List<BoostedNode> node1bPredecessor = BoostedNodeUtils
		// .getPredecessors(node1b);
		// List<BoostedNode> node1bSuccessor = BoostedNodeUtils
		// .getSuccessors(node1b);
		// // getLog().info(Refactor.getStructureString(node1b));
		//
		//
		// BoostedNode node2a = conflict2.getValue().getLeft();
		// // List<BoostedNode>node2aPredecessor =
		// // BoostedNodeUtils.getPredecessors(node2a);
		// // List<BoostedNode>node2aSuccessor =
		// // BoostedNodeUtils.getSuccessors(node2a);
		// // getLog().info(Refactor.getStructureString(node2a));
		//
		// BoostedNode node2b = conflict2.getValue().getRight();
		// // List<BoostedNode>node2bPredecessor =
		// // BoostedNodeUtils.getPredecessors(node2b);
		// // List<BoostedNode>node2bSuccessor =
		// // BoostedNodeUtils.getSuccessors(node2b);
		// // getLog().info(Refactor.getStructureString(node2b));
		//
		// boolean node1aIsPredecessorOfNode2a = node1aSuccessor
		// .contains(node2a);
		// boolean node1aIsPredecessorOfNode2b = node1aSuccessor
		// .contains(node2b);
		// boolean node1bIsPredecessorOfNode2a = node1bSuccessor
		// .contains(node2a);
		// boolean node1bIsPredecessorOfNode2b = node1bSuccessor
		// .contains(node2b);
		//
		// boolean node1aIsSuccessorOfNode2a = node1aPredecessor
		// .contains(node2a);
		// boolean node1aIsSuccessorOfNode2b = node1aPredecessor
		// .contains(node2b);
		// boolean node1bIsSuccessorOfNode2a = node1bPredecessor
		// .contains(node2a);
		// boolean node1bIsSuccessorOfNode2b = node1bPredecessor
		// .contains(node2b);
		//
		// if (node1aIsSuccessorOfNode2a
		// && node1aIsPredecessorOfNode2b)
		// return 0;
		// if (node1aIsSuccessorOfNode2b
		// && node1aIsPredecessorOfNode2a)
		// return 0;
		// if (node1bIsSuccessorOfNode2a
		// && node1bIsPredecessorOfNode2b)
		// return 0;
		// if (node1bIsSuccessorOfNode2b
		// && node1bIsPredecessorOfNode2a)
		// return 0;
		// if (node1aIsPredecessorOfNode2a
		// && node1bIsSuccessorOfNode2a)
		// return 0;
		// if (node1aIsPredecessorOfNode2b
		// && node1bIsSuccessorOfNode2b)
		// return 0;
		// if (node1bIsPredecessorOfNode2a
		// && node1aIsSuccessorOfNode2a)
		// return 0;
		// if (node1bIsPredecessorOfNode2b
		// && node1aIsSuccessorOfNode2b)
		// return 0;
		//
		// if (node1aIsPredecessorOfNode2a) {
		// if (node1bIsSuccessorOfNode2b) {
		// return 0;
		// } else {
		// return -1;
		// }
		// }
		//
		// if (node1aIsPredecessorOfNode2b) {
		// if (node1bIsSuccessorOfNode2a) {
		// return 0;
		// } else {
		// return -1;
		// }
		// }
		//
		// if (node1bIsPredecessorOfNode2a) {
		// if (node1aIsSuccessorOfNode2b) {
		// return 0;
		// } else {
		// return -1;
		// }
		// }
		//
		// if (node1bIsPredecessorOfNode2b) {
		// if (node1aIsSuccessorOfNode2a) {
		// return 0;
		// } else {
		// return -1;
		// }
		// }
		//
		// if (node1aIsSuccessorOfNode2a || node1aIsSuccessorOfNode2b
		// || node1bIsSuccessorOfNode2a
		// || node1bIsSuccessorOfNode2b) {
		// return 1;
		// }
		//
		// return 0;
		// } catch (Exception ex) {
		// throw ThrowableBoostUtils.toRuntimeException(ex);
		// }
		// }
		// };
		// simpleDisplay(10, true);
		//
		// Refactor.findComparatorProblems(comparator, result);
		//
		// Collections.sort(result, comparator);
		// return result;

	}

	/**
	 * A plan is solved if all operations are elementary, no conflicts occur, no virtual objects are used and all conditions are fulfilled.
	 * 
	 * @return true if the plan is solved
	 */
	public boolean isSolved() throws Exception {
		// getLog().error("isSolved does not work correctly. Bad initialization");
		boolean result = !(hasConflicts() || hasUnsolvedNodes()
				|| hasVirtualObjects() || hasGoalConflicts());
		return result;
	}

	/**
	 * Are there any (possible) conflicts in the plan?
	 * 
	 * @return true if possible conflicts are found
	 */
	public boolean hasConflicts() {
		// getLog().error("implement hasConflicts");

		return !possibleConflicts.isEmpty();
	}

	/**
	 * Are there any nodes that need to be processed/solved.
	 * 
	 * @return true if there is at least one unsolved node
	 */
	public boolean hasUnsolvedNodes() throws Exception {
		//getLog().error("implement hasUnsolvedNodes");
		return !getUnsolvedNodes().isEmpty();
		
	}

	/**
	 * Are there still any virtual objects present that need to be replaced with instances existing in the world
	 * 
	 * @return true if there is at least one virtual object
	 */
	public boolean hasVirtualObjects() {
		return !getNodesWithVirtualProperties().isEmpty();

	}


	/**
	 * Extract the operation of a node
	 * 
	 * @param node
	 *            the node to be inspected
	 * @return the operation of the node
	 */
	public Operation<Conditio, NameOfObjects> getOperation(BoostedNode node) {
		return (Operation<Conditio, NameOfObjects>) node.getAttributes()
				.getValue();

	}

	/**
	 * Are there any goal conflicts in the plan
	 * 
	 * @return true if at least one conflict is found.
	 */
	public boolean hasGoalConflicts() throws Exception {

		Iterator<BoostedNode> goalIterator = graph.getAllSinkVertices()
				.iterator();

		while (goalIterator.hasNext() ) {
			Operation<Conditio, NameOfObjects> goal = getOperation(goalIterator
					.next());
			if (world.isToSolveOperator(goal) && world.isNoNeedToSolve(goal)
					&& goal.getPostcondition() != null) {
				for (BoostedNode otherGoalNode : graph.getAllSinkVertices()) {
					Operation<Conditio, NameOfObjects> otherGoal = getOperation(otherGoalNode);
					if (otherGoal != goal && otherGoal.getGlobalPostcondition() != null){
						if (goal.getPostcondition().isInConflict(otherGoal.getGlobalPostcondition()))
							return true;
					}
				}
			}
		}

		return false;


	}

	/**
	 * Remove operators that are superfluous. Superfluous operators are
	 * operators that do nothing. These are operators which which global
	 * postcondition is already implied by the global precondition. The global
	 * postcondition is used and not the postcondition, because the
	 * postcondition may not be defined exactly enough to determine whether
	 * anything has been done by the operator
	 * 
	 * Operators are only removed, when they are not in conflict with another
	 * operator or the plan is already solved
	 * 
	 * @return true if the graph has been changed
	 */
	public boolean removeSuperfluousOperators() throws Exception {

		selfcheck();
		boolean result = false;
		for (BoostedNode node : graph.getAllVertices()) {
			Operation operation = getOperation(node);
			Condition globalPrecondition = operation.getGlobalPrecondition();
			Condition globalPostCondition = operation.getGlobalPostcondition();
			if (globalPrecondition != null && globalPostCondition != null
					&& (isSolved() || getConflictsWith(node).isEmpty())) {
				if (globalPrecondition.implies(globalPostCondition)) {
					graph.removeNode(node);

					result = true;
				}
			}
		}
		// getLog().error("implement removeSuperfluousOperators");
		selfcheck();
		return result;


	}


	/**
	 * Creates one to-solve-node
	 * 
	 * @param globalPrecondition
	 *            the global precondition of the node
	 * @param postcondition
	 *            the postcondition of the node
	 */
	public void initialize(World<Conditio, NameOfObjects> world,
			Conditio globalPrecondition,
			Conditio postcondition) throws Exception {
		
		Operation<Conditio, NameOfObjects> riddle = Operation
				.createProblemToSolveNode(world,
				globalPrecondition,
				postcondition);
		initialize(world, riddle);

	}

	/**
	 * Create one node
	 * 
	 * @param riddle
	 *            the operation of the node (the riddle to solve)
	 */
	public void initialize(World<Conditio, NameOfObjects> world,
			Operation<Conditio, NameOfObjects> riddle) throws Exception {
		riddle.checkOperator();
		this.world = world;
		setWorldInstanceDescription(riddle.getGlobalPrecondition());

		BoostedNode riddleNode = createNode(riddle);
		graph = new BoostedNodeGraph(riddleNode);
		graph.setAtMost1EdgeBetweenNodes(true);

		possibleConflicts = new HashSet<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>>();
		world.getComputeGlobalPostconditionListener().add(this);
	}


	/**
	 * Create one node.
	 * 
	 * @param content
	 *            the operation of the node
	 * @return the created node
	 */
	public BoostedNode createNode(Operation<Conditio, NameOfObjects> content)
			throws Exception {
		SimpleBoostedNode<String> result = new SimpleBoostedNode<String>();
		result.getAttributes().setValue(content);
		result.setName("" + content.getName());
		return result;
	}

	/**
	 * Do a simple display of the plan
	 */
	public void simpleDisplay() {
		simpleDisplay(getSolveStateDisplayConfiguration());
	}

	/**
	 * Do a simple display of the plan
	 * 
	 * @param displayConfiguration
	 *            the configuration of the display
	 */
	public void simpleDisplay(
			GraphDisplayConfiguration<BoostedNode, NameOfObjects> displayConfiguration) {

		displayConfiguration.setTitle("Plan "
				+ ((version != null) ? " (" + version + ")" : ""));
		graph.simpleDisplay(displayConfiguration);

	}

	public Integer getVersion() {
		return version;
	}

	public void setVersion(Integer version) {
		this.version = version;
	}

	public World<Conditio, NameOfObjects> getWorld() {
		return world;
	}

	public Set<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>> getPossibleConflicts() {
		return possibleConflicts;
	}

	public void setPossibleConflicts(
			Set<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>> possibleConflicts) {
		this.possibleConflicts = possibleConflicts;
	}

	/**
	 * Find all unsolved nodes of this plan
	 * 
	 * @return all unsolved nodes
	 */
	public Set<BoostedNode> getUnsolvedNodes() throws Exception {
		HashSet<BoostedNode> result = new HashSet<BoostedNode>();
		for (BoostedNode node : graph.getAllVertices()) {
			Operation<Conditio, NameOfObjects> operation = Operation
					.getOperation(node);
			if (!(operation.getOperator().isElementary()
					|| operation.isNoNeedToSolve() || operation.isSolved())) {
				result.add(node);
			}
		}
		return result;
	}

	/**
	 * Transforms the unsolved nodes
	 * 
	 * @param transformations
	 *            with signature Collection<Operation>
	 *            getObjectInstance(Collection<Operation>, ...)
	 * @return the transformed nodes
	 */
	@SuppressWarnings("unchecked")
	public Collection<BoostedNode> getUnsolvedNodes(
			ObjectFactory... transformations) throws Exception {
		return (Collection<BoostedNode>) TransformationGraph.createChain(
				transformations).getObjectInstance(getUnsolvedNodes(), null,
				null, null);
	}

	/**
	 * Find all nodes that have virtual properties
	 * 
	 * @return all found nodes
	 */
	public Set<BoostedNode> getNodesWithVirtualProperties() {
		HashSet<BoostedNode> result = new HashSet<BoostedNode>();
		for (BoostedNode node : graph.getAllVertices()) {
			if (world.hasVirtualProperty(node)) {
				result.add(node);
			}
		}
		return result;
	}


	public void setWorld(World<Conditio, NameOfObjects> world) {
		this.world = world;
	}

	public BoostedNodeGraph getGraph() {
		return graph;
	}

	/**
	 * Replaces an node by a subgraph. The global conditions etc. are adapted.
	 * 
	 * @param nodeOfPlan
	 *            the node that is to be replaced
	 * @param replacement
	 *            the replacement of the node
	 */
	public void replaceNode(BoostedNode nodeOfPlan, BoostedNodeGraph replacement)
			throws Exception {

		// transfer global preconditions
		Operation<Conditio, NameOfObjects> unsolvedPlanNode = Operation
				.getOperation(nodeOfPlan);
		Conditio globalPrecondition = unsolvedPlanNode.getGlobalPrecondition();
		for (BoostedNode sourceNode : replacement.getAllSourceVertices()) {
			Operation<Conditio, NameOfObjects> operation = Operation
					.getOperation(sourceNode);
			operation.setGlobalPrecondition(globalPrecondition);
		}


		graph.replace(nodeOfPlan, replacement);

		nodesUpdated(replacement.getAllVertices(), Arrays.asList(nodeOfPlan));

	}

	/**
	 * Informs that some nodes have been updated. Redundancies of the graph edges are removed. Conditions and conflicts are updated.
	 * 
	 * @param updatedNodes
	 *            the nodes that were updated
	 * @param removedNodes
	 *            the nodes that were removed
	 */
	private void nodesUpdated(Collection<BoostedNode> updatedNodes,
			Collection<BoostedNode> removedNodes) throws Exception {
		graph.removeRedundancies();
		updateGlobalConditions(updatedNodes, removedNodes);
		updatePossibleConflicts(updatedNodes, removedNodes);

	}

	/**
	 * Update the data about possible conflicts
	 */
	private void updatePossibleConflicts(Collection<BoostedNode> updatedNodes,
			Collection<BoostedNode> removedNodes) throws Exception {

		// first remove all existing conflicts that refer to the nodes
		removeFromPossibleConflicts(updatedNodes);
		removeFromPossibleConflicts(removedNodes);

		addPossibleConflictsOf(updatedNodes);
	}

	/**
	 * Add new possible conflicts
	 * 
	 * @param nodesThatMayHaveConflicts
	 *            inspect these nodes and check wether there are conflicts.
	 */
	private void addPossibleConflictsOf(
			Collection<BoostedNode> nodesThatMayHaveConflicts) {

		addPossibleConflictsBetweenParallelNodes(nodesThatMayHaveConflicts);

		// add other conflicts

	}

	/**
	 * Adds all conflicts with nodes parallel to some given nodes.
	 * 
	 * @param nodesThatMayHaveConflicts
	 *            the nodes that may be in conflict
	 */
	private void addPossibleConflictsBetweenParallelNodes(
			Collection<BoostedNode> nodesThatMayHaveConflicts) {

		Set<Pair<BoostedNode, BoostedNode>> parallelNodePairsToInspect = graph
				.getParallelNodesOfForest(nodesThatMayHaveConflicts);
		for (Pair<BoostedNode, BoostedNode> nodes : parallelNodePairsToInspect) {
			addPossibleConflictsOf(nodes.getLeft(), nodes.getRight());
		}

	}

	/**
	 * Compute whether there are conflicts between the two nodes. The structure
	 * of the graph is not considered
	 * 
	 */
	private void addPossibleConflictsOf(BoostedNode node1, BoostedNode node2) {
		addPossiblePropertyConflictsOf(node1, node2);
		// other conflicts may follow here
	}

	/**
	 * Add the possible property conflicts of two nodes to the conflicts list
	 * 
	 * @param node1
	 * @param node2
	 */
	private void addPossiblePropertyConflictsOf(BoostedNode node1,
			BoostedNode node2) {
		for (Pair<String, String> conflictingProperties : world
				.getPossiblePropertyConflictsDefinition()) {

			Operation<Conditio, NameOfObjects> operation1 = getOperation(node1);
			Map<String, Set<String>> properties1 = world
					.getPropertyValues(operation1);
			Set<String> firstProperty1 = properties1.get(conflictingProperties
					.getLeft());
			Set<String> secondProperty1 = properties1.get(conflictingProperties
					.getRight());

			Operation<Conditio, NameOfObjects> operation2 = getOperation(node2);
			Map<String, Set<String>> properties2 = world
					.getPropertyValues(operation2);
			Set<String> firstProperty2 = properties2.get(conflictingProperties
					.getLeft());
			Set<String> secondProperty2 = properties2.get(conflictingProperties
					.getRight());

			if (firstProperty1 != null && secondProperty2 != null) {
				HashSet<String> intersection = new HashSet<String>(
						firstProperty1);
				intersection.retainAll(secondProperty2);
				if (!intersection.isEmpty()) {
					ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>> conflict = new ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>();
					conflict.setType("property conflict");
					conflict.setName(conflictingProperties.getLeft() + "-"
							+ conflictingProperties.getRight() + ": "
							+ intersection);
					conflict.setValue(new ImmutablePair<BoostedNode, BoostedNode>(
							node1, node2));
					possibleConflicts.add(conflict);
					return;
				}
			}

			if (firstProperty2 != null && secondProperty1 != null) {
				HashSet<String> intersection = new HashSet<String>(
						firstProperty2);
				intersection.retainAll(secondProperty1);
				if (!intersection.isEmpty()) {
					ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>> conflict = new ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>();
					conflict.setType("property conflict");
					conflict.setName(conflictingProperties.getRight() + "-"
							+ conflictingProperties.getLeft() + ": "
							+ intersection);
					conflict.setValue(new ImmutablePair<BoostedNode, BoostedNode>(
							node1, node2));
					possibleConflicts.add(conflict);
					return;
				}
			}

		}

	}

	/**
	 * Removes all conflicts concerning some given nodes
	 * 
	 * @param nodes
	 *            the nodes which conflicts are to be removed.
	 */
	private void removeFromPossibleConflicts(Collection<BoostedNode> nodes) {
		for (ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>> conflict : new ArrayList<ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>>>(
				possibleConflicts)) {
			// use a new array list, so that the elements can be removed
			// instantaneously
			if (nodes.contains(conflict.getValue().getLeft())
					|| nodes.contains(conflict.getValue().getRight())) {
				possibleConflicts.remove(conflict);
			}
		}
	}


	/**
	 * Updates the global conditions of the nodes
	 * 
	 * @return true if a null condition changed to a non-null condition
	 */
	public boolean updateGlobalConditions() throws Exception {

		return updateGlobalConditions(graph.getAllVertices(),
				Collections.EMPTY_LIST);
	}

	/**
	 * Updates the global conditions of some nodes
	 * 
	 * @return true if a null condition changed to a non-null condition
	 */
	private boolean updateGlobalConditions(
			Collection<BoostedNode> updatedNodes,
			Collection<BoostedNode> removedNodes) throws Exception {

		boolean result = false;
		Set<BoostedNode> sourceVertices = graph.getAllSourceVertices();

		boolean doDisplay = false;
		for (BoostedNode node : graph.getAllVerticesOrdered()) {
			// if (doDisplay) {
			// simpleDisplay(getSolveStateDisplayConfiguration());
			// }

			Operation<Conditio, NameOfObjects> operation = getOperation(node);


			// update global precondition
			if (sourceVertices.contains(node)) {
				// nothing to do
			} else {
				// the new global precondition is the combination of all global
				// postconditions of the predecessors
				Conditio newGlobalPrecondition = null;
				for (BoostedNode predecessor : BoostedNodeUtils
						.getPredecessors(node)) {
					Operation<Conditio, NameOfObjects> predecessorOperation = getOperation(predecessor);

					if (predecessorOperation.getGlobalPostcondition() == null) {
						newGlobalPrecondition = null;
						break;
						// because the new global postcondition can not be
						// computed.
					} else {

						Conditio clonedPostcondition = predecessorOperation
								.getGlobalPostcondition().cloneMe();
						if (newGlobalPrecondition == null) {
							newGlobalPrecondition = clonedPostcondition;
						} else {
							newGlobalPrecondition.add(clonedPostcondition);
						}
					}
				}
				if (operation.getGlobalPrecondition() == null) {
					result = true;
				}
				operation.setGlobalPrecondition(newGlobalPrecondition);
			}

			Conditio globalPostcondition = operation.getGlobalPostcondition();

			warningHappened = false;
			errorHappened = false;
			fatalErrorHappened = false;
			
			world.updateGlobalPostcondition(operation);

			if (warningHappened || errorHappened || fatalErrorHappened) {
				// graph.simpleDisplay();
				// List<BoostedNode> sequence = graph.getSequenceBetween(graph
				// .getAllSourceVertices().iterator().next(), node);
				// for (BoostedNode bn : sequence) {
				// System.out.println(getOperation(bn).toString(3));
				// }
				// System.out.println("remove this code later");
				getLog().info("warning or error or fatal error happened");
			}

			if (globalPostcondition == null
					&& operation.getGlobalPostcondition() != null) {
				result = true;
			}

		}

		// if (doDisplay) {
		// simpleDisplay(getSolveStateDisplayConfiguration());
		// }

		return result;
	}


	/**
	 * Replaces one virtual objcts with a real instance of the world
	 * 
	 * @param nodeWithVirtualObject
	 *            the node that contains the virtual object
	 * @param virtualObject
	 *            the virtual object to be replaced
	 * @param concreteObject
	 *            the replacement
	 */
	public void instantiateVirtualObject(BoostedNodeGraph graph,
			BoostedNode nodeWithVirtualObject,
			WorldObject<NameOfObjects, ?> virtualObject,
			HasName<NameOfObjects> concreteObject) throws Exception {

		// clone the virtual object, because it may be changed during the
		// instantiation
		virtualObject = virtualObject.cloneMe();

		BoostedNode newVirtualObjectNode = nodeWithVirtualObject;
		Set<BoostedNode> changedNodes = new HashSet<BoostedNode>();
		// for each successor of the node until the delete node: replace the
		// name of the virtual object with the name of the concrete object

		// first find all affected nodes without changing anything (otherwise
		// there may be problems in finding all relevant nodes)
		BoostedNode deleteVirtualObjectNode = findVirtualObjectNodes(
				changedNodes, nodeWithVirtualObject, virtualObject, null);

		NameOfObjects newName = concreteObject.getName();
		for (BoostedNode node : changedNodes) {
			Operation<Conditio, NameOfObjects> operation = Operation
					.getOperation(node);
			operation.replaceObjectName(virtualObject, newName);
		}

		// remove the new and delete node of the virtual object
		if (deleteVirtualObjectNode == null) {
			throw new IllegalStateException(
					"No delete virtual object node has been found. This can not be.");
		}

		graph.removeNode(newVirtualObjectNode);
		graph.removeNode(deleteVirtualObjectNode);

		changedNodes.remove(newVirtualObjectNode);
		changedNodes.remove(deleteVirtualObjectNode);

		// check
		for (BoostedNode node : graph.getAllVertices()) {
			Operation/*<?, ?>*/ operation = Operation.getOperation(node);
			HasName<?> operatorDetails = operation.getOperatorDetails();
			Check.notNull(virtualObject, "virtualObject");
			if (operatorDetails != null) {
				if (virtualObject.getName().equals(operatorDetails.getName())) {
					throw new IllegalStateException(
							"virtual object not replaced correctly");
				}
			}
		}

		nodesUpdated(changedNodes,
				Arrays.asList(newVirtualObjectNode, deleteVirtualObjectNode));

		selfcheck();
	}

	/**
	 * Find nodes that contain a given virtual object.
	 * 
	 * @param alreadyInspected
	 *            the nodes that have already been inspected
	 * @param toInspect
	 *            the node to be inspected first
	 * @param virtualObject
	 *            the virtual object to search for
	 * @param deleteVirtualObjectNode
	 *            the node that contains the delete-operation of the virtual object (may be null, if the node is unknown)
	 * @return the delete virtual object node (if found)
	 */
	private BoostedNode findVirtualObjectNodes(
			Set<BoostedNode> alreadyInspected, BoostedNode toInspect,
			HasName virtualObject, BoostedNode deleteVirtualObjectNode) {
		
		if ( ! alreadyInspected.contains(toInspect)){
			alreadyInspected.add(toInspect);
			Operation/*<?, ?>*/ operation = Operation.getOperation(toInspect);
			boolean isDeleteNode = operation.getOperator().getName()
					.equals(World.OPERATORNAME_DELETE_VIRTUAL_OBJECT)
					&& virtualObject.equals(operation.getOperatorDetails());

			if (isDeleteNode) {
				deleteVirtualObjectNode = toInspect;
				//no successors need to be inspected
			} else {
				//inspect successors
				for (BoostedNode successor: BoostedNodeUtils.getSuccessors(toInspect)){
					deleteVirtualObjectNode = findVirtualObjectNodes(
							alreadyInspected, successor, virtualObject,
							deleteVirtualObjectNode);
				}
			}
			
		}
		
		return deleteVirtualObjectNode;
	}

	/**
	 * Gets a clone of the description.
	 * 
	 */
	public Conditio getWorldInstanceDescription() throws Exception {
		// cloning to prevent changes because of side effects
		return worldInstanceDescription.cloneMe();
	}

	public void setWorldInstanceDescription(Conditio worldInstanceDescription)
			throws Exception {
		// cloning to prevent changes because of side effects
		this.worldInstanceDescription = worldInstanceDescription.cloneMe();
	}

	/**
	 * Adds a precondition to the source nodes of a graph
	 * 
	 * @param subgraph
	 *            the inspected graph
	 * @param additionalPrecondition
	 *            the precondition to be added
	 */
	public void addPreconditionToSourceNodes(BoostedNodeGraph subgraph,
			Conditio additionalPrecondition) throws Exception {
		addPreconditionToNodes(subgraph.getAllSourceVertices(),
				additionalPrecondition);

	}

	/**
	 * Adds a precondition to some nodes
	 * 
	 * @param nodes
	 *            will be modified
	 * @param additionalPrecondition
	 *            the precondition to be added
	 */
	public void addPreconditionToNodes(Collection<BoostedNode> nodes,
			Conditio additionalPrecondition) throws Exception {
		for (BoostedNode sourceNode : nodes) {
			Operation<Conditio, NameOfObjects> operation = Operation
					.getOperation(sourceNode);
			operation.addToPrecondition(additionalPrecondition);
		}

	}


	/**
	 * Adds an additional global precondition to the source nodes of a graph
	 * 
	 * @param subgraph
	 *            contains the source nodes
	 */
	public void addGlobalPreconditionToSourceNodes(BoostedNodeGraph subgraph,
			Conditio additionalPrecondition) throws Exception {
		addGlobalPreconditionToNodes(subgraph.getAllSourceVertices(),
				additionalPrecondition);
	}

	/**
	 * Adds an additional global precondition to some nodes
	 * 
	 * @param nodes
	 *            will be modified
	 */
	public void addGlobalPreconditionToNodes(Collection<BoostedNode> nodes,
			Conditio additionalPrecondition) throws Exception {
		for (BoostedNode sourceNode : nodes) {
			Operation<Conditio, NameOfObjects> operation = Operation
					.getOperation(sourceNode);
			operation.addToGlobalPrecondition(additionalPrecondition);
		}

	}

	/**
	 * Adds the default conditions of the operations to the nodes of the subgraph
	 * 
	 * @param subgraph
	 *            its nodes will be modified
	 */
	public void addDefaultConditionsOfOperators(BoostedNodeGraph subgraph)
			throws Exception {
		for (BoostedNode node : subgraph.getAllVertices()) {
			Operation<Conditio, NameOfObjects> operation = Operation
					.getOperation(node);
			operation.addToPostcondition(world.getDefaultPostcondition(this,
					operation));
			operation.addToDeletecondition(world.getDefaultDeletecondition(
					this, operation));
			operation.addToPrecondition(world.getDefaultPrecondition(this,
					operation));

		}
	}

	/**
	 * Gets all nodes that have equal operation
	 * 
	 * @return a set A containing sets B of nodes. Each BoostedNode of set B has
	 *         the same operation
	 */
	public Set<Set<BoostedNode>> getNodesWithEqualOperations() {

		HashSet<Set<BoostedNode>> result = new HashSet<Set<BoostedNode>>();

		HashMap<Pair<Operator, WorldObject>, Set<BoostedNode>> map = new HashMap<Pair<Operator, WorldObject>, Set<BoostedNode>>();

		for (BoostedNode node : graph.getAllVertices()) {
			Operation<Conditio, NameOfObjects> operation = getOperation(node);
			Refactor.putIntoMapOfSets(
					map,
					new ImmutablePair<Operator, WorldObject>(operation
							.getOperator(), operation.getOperatorDetails()),
					node);
		}

		// remove empty sets and sets with only one element
		// use a new collection, to be able to remove without problems
		for (Set<BoostedNode> set : new ArrayList<Set<BoostedNode>>(result)) {
			if (set.isEmpty() || set.size() == 1) {
				result.remove(set);
			}
		}
		return result;

	}

	/**
	 * Try to combine nodes with equal operators, be changing the structure of the plan. If nodes do the same operation in parallel branches, it may be possible
	 * merge the two operations into one by changing the plan structure. One operation is removed, its predecessors become predecessors of the other operation;
	 * its successors become successors of the other operation.
	 * 
	 * @param nodesWithEqualsOperators
	 *            the inspeced nodes
	 * @return true if something was changed
	 */
	public boolean combineNodesWithEqualOperators(
			BoostedNode... nodesWithEqualsOperators) {
		throw new IllegalStateException(
				"combineNodesWithEqualOperators not yet implemented");
		// updateGlobalConditions();
		
	}

	/**
	 * Finds all nodes that are in conflict with a given node
	 * 
	 * @param node
	 *            the inspected node
	 * @return the conflicting nodes
	 */
	private Set<BoostedNode> getConflictsWith(BoostedNode node){
		HashSet<BoostedNode> result = new HashSet<BoostedNode>();
		for (ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>> conflict: possibleConflicts){
			if (conflict.getValue().getLeft().equals(node)){
				result.add(conflict.getValue().getRight());
			} else if (conflict.getValue().getRight().equals(node)){
				result.add(conflict.getValue().getLeft());
			}
		}
		
		return result;
	}
	
	/**
	 * Starting with a startnode this algorithm follows the serial sequence (at most up to the next split). It returns the last node that is in conflict with a
	 * given conflictNode.
	 * 
	 * @param startnode
	 *            the node to start with
	 * @param conflictNode
	 *            the node which conflicts are inspeced
	 * @return the last node of the sequence with a conflict with the node
	 */
	private BoostedNode getLastConflictInASeries(BoostedNode startnode, BoostedNode conflictNode)
	{
		Set<BoostedNode> conflictingNodes = getConflictsWith(conflictNode);
		BoostedNode endnode = startnode;
		boolean doContinue = true;
		while (doContinue){
			List<BoostedNode> successors = BoostedNodeUtils.getSuccessors(endnode);
			if (successors.size() != 1){
				//stop if no successor exists
				//stop if more than one successors exist, because then you are leaving the sequence
				doContinue = false;
			} else {
				BoostedNode successor = successors.get(0);
				if (conflictingNodes.contains(successor)){
					endnode = successor;
				} else {
					doContinue = false;
				}
			}
		}
		return endnode;
	}
	
	/**
	 * Do a simple display.
	 * 
	 * @param colors
	 *            colors of the nodes
	 */
	private void display(Map<BoostedNode, Color> colors) {
		GraphDisplayConfiguration<BoostedNode, String> display = GraphDisplayConfiguration
				.createFilledNodesWithoutTextConfiguration();
		display.setVertexColors(colors);
		graph.simpleDisplay(display);

	}

	/**
	 * Try to solve a possible conflict be ordering the affected nodes
	 * 
	 * @param conflict
	 *            the conflicting nodes
	 * 
	 * @return true if something was changed
	 */
	public boolean solvePossibleConflictByReordering(
			ObjectAttributes<String, String, Pair<BoostedNode, BoostedNode>> conflict)
			throws Exception {

		BoostedNode node1 = conflict.getValue().getLeft();
		BoostedNode node2 = conflict.getValue().getRight();

		if (true) {
			display(ContainerBoostUtils.createMap(node1, Color.RED, node2,
					Color.GREEN));
		}

		if (graph.areConnected(node1, node2)) {
			// the conflict may be no longer be a conflict because the structure
			// changed.
			getLog().trace(
					"the conflicting nodes are not (longer) parallel, removing the conflict");
			possibleConflicts.remove(conflict);
			return true;
		} else {
			BoostedNode endnode1 = getLastConflictInASeries(node1, node2);
			BoostedNode endnode2 = getLastConflictInASeries(node2, node1);

			if (true) {
				display(ContainerBoostUtils.createMap(endnode1, Color.ORANGE,
						endnode2, Color.CYAN, node1, Color.RED, node2,
						Color.GREEN));
			}
			
			if (graph.areConnected(endnode1, endnode2)) {
				throw new IllegalStateException("the endnodes should not be connected, because the nodes were not connected");
			}

			Set<BoostedNode> modifiedNodes = serializeNodeSeries(node1, endnode1, node2, endnode2);
			nodesUpdated(modifiedNodes, Collections.EMPTY_SET);
			return !modifiedNodes.isEmpty();
			
		}



	}
	
	/**
	 * Serializes two sequences of nodes
	 * 
	 * @param startnode1
	 *            the start of the first sequence
	 * @param endnode1
	 *            the end of the first sequence
	 * @param startnode2
	 *            the start of the second sequence
	 * @param endnode2
	 *            the end of the second sequence
	 * @return the modified nodes
	 */
	private Set<BoostedNode> serializeNodeSeries(BoostedNode startnode1,
			BoostedNode endnode1, BoostedNode startnode2, BoostedNode endnode2)
			throws Exception {
		
		// TODO save structure if a problem exists
		
		if (false) {
			display(ContainerBoostUtils.createMap(startnode1, Color.RED,
					endnode1, Color.RED, startnode2, Color.CYAN, endnode2,
					Color.cyan));
		}

		Set<BoostedNode> modified = new HashSet<BoostedNode>();

		boolean result = serializeNode2BeforeNode1(startnode1, endnode2);
		if (!result) {
			result = serializeNode2BeforeNode1(startnode2, endnode1);
		}
		if (!result) {
			result = serializeNode2AfterNode1(endnode1, startnode2);
		}
		if (!result) {
			result = serializeNode2AfterNode1(endnode2, startnode1);
		}

		if (result) {
			modified.addAll(graph.getSequenceBetween(
					startnode1, endnode1));
			modified.addAll(graph.getSequenceBetween(startnode2, endnode2));
		}
		// graph.simpleDisplay();
		return modified;
	}

	/**
	 * Serialize by reordering, so that node2 is before node1
	 * 
	 * @return true if something was changed
	 */
	private boolean serializeNode2BeforeNode1(BoostedNode node1,
			BoostedNode node2) throws Exception {

		// GraphDisplayConfiguration<BoostedNode, String> display = new
		// GraphDisplayConfiguration<BoostedNode, String>();
		// display.setHighlightedVertices(Arrays.asList(node1, node2));
		// graph.simpleDisplay(display);
		// predecessors of node 1 (including node 1)
		Queue<BoostedNode> predecessors = new ConcurrentLinkedQueue<BoostedNode>();
		predecessors.add(node1);
		BoostedNode predecessor;
		Conditio postcondition2 = getOperation(node2).getPostcondition();
		boolean result = false;
		Set<BoostedNode> handledPredecessors = new HashSet<BoostedNode>();
		while ((!predecessors.isEmpty()) && !result) {
			predecessor = predecessors.remove();
			if (handledPredecessors.contains(predecessor)) {
				// predecessor has already been handled. No need to do it again
				// throw new IllegalStateException(
				// "there may be a problem. Need better prove");

			} else {

				Conditio precondition1 = getOperation(predecessor)
						.getPrecondition();
				if (ConditionBasics.isInConflict(postcondition2, precondition1)) {
					predecessors.addAll(BoostedNodeUtils
							.getPredecessors(predecessor));
				} else {
					if (graph.areConnected(node2, predecessor)) {
						// serialization is no longer possible, because the
						// nodes
						// are not parallel
					} else {
						boolean isSink = BoostedNodeUtils.getSuccessors(node2)
								.isEmpty();
						Refactor.addEdge(node2, predecessor);
						if (isSink) {
							// // insert a new goal-node with the same
							// postcondition
							// as
							// node2
							Operation<Conditio, NameOfObjects> newSink = new Operation<Conditio, NameOfObjects>();
							newSink.setOperator(getWorld().getToSolveOperator());
							newSink.setPostcondition(postcondition2.cloneMe());
							BoostedNode newSinkNode = graph
									.addNodeWithValue(newSink);
							Refactor.addEdge(node2, newSinkNode);
						}
					}
				}
			}
		}

		return result;
	}

	/**
	 * Serialize by reordering, so that node2 is after node1
	 * 
	 * @return true if something was changed
	 */
	private boolean serializeNode2AfterNode1(BoostedNode node1,
			BoostedNode node2) throws Exception {
		// predecessors of node 1 (including node 1)
		Queue<BoostedNode> successors = new ConcurrentLinkedQueue<BoostedNode>();
		successors.add(node1);
		BoostedNode successor;
		Conditio precondition2 = getOperation(node2).getPrecondition();
		boolean result = false;
		Set<BoostedNode> handledSuccessors = new HashSet<BoostedNode>();
		while ((!successors.isEmpty()) && !result) {
			successor = successors.remove();
			if (handledSuccessors.contains(successor)) {
				// successor has already been handled. No need to do it again
				// throw new IllegalStateException(
				// "there may be a problem. Need better prove");
			} else {
				handledSuccessors.add(successor);
				Conditio postcondition1 = getOperation(successor)
						.getPostcondition();
				if (ConditionBasics.isInConflict(postcondition1, precondition2)) {
					successors
							.addAll(BoostedNodeUtils.getSuccessors(successor));
				} else {
					if (graph.areConnected(node2, successor)) {
						// serialization is no longer possible, because the
						// nodes
						// are not parallel
					} else {
						boolean isSink = BoostedNodeUtils.getSuccessors(
								successor).isEmpty();
						Refactor.addEdge(successor, node2);
						if (isSink) {
							// // insert a new goal-node with the same
							// postcondition
							// as
							// node2
							Operation<Conditio, NameOfObjects> newSink = new Operation<Conditio, NameOfObjects>();
							newSink.setOperator(getWorld().getToSolveOperator());
							newSink.setPostcondition(postcondition1.cloneMe());
							BoostedNode newSinkNode = graph
									.addNodeWithValue(newSink);
							Refactor.addEdge(successor, newSinkNode);
						}
					}
				}
			}
		}

		return result;

	}


	/**
	 * Create a display configuration that colors the nodes according to their solve-state (unsolved, solved, ...)
	 * 
	 * @return the created configuration
	 */
	public GraphDisplayConfiguration<BoostedNode, NameOfObjects> getSolveStateDisplayConfiguration() {
		GraphDisplayConfiguration<BoostedNode, NameOfObjects> displayConfiguration = GraphDisplayConfiguration
				.createFilledNodesWithoutTextConfiguration();
		displayConfiguration.setDisplayVertexValueInsteadOfName(true);
		displayConfiguration.setVertexDetailDisplayLevel(10);
		displayConfiguration
				.setVertexBackgroundTransformer(new Transformer<BoostedNode, Paint>() {
					@Override
					public Paint transform(BoostedNode node) {
						Color result = null;
						try {
							Operation<Conditio, NameOfObjects> operation = getOperation(node);
							if (operation.isSolved()) {
								result = Color.GREEN;
							} else if (operation.isNoNeedToSolve()) {
								result = Color.CYAN;
							} else {
								Conditio globalPrecondition = operation
										.getGlobalPrecondition();
								Conditio precondition = operation
										.getPrecondition();
								if (globalPrecondition != null
										&& precondition != null) {
									if (!globalPrecondition
											.implies(precondition)) {
										result = Color.RED;
									}
								}

								if (result == null) {
									if (operation.getGlobalPostcondition() != null) {
										result = Color.YELLOW;
									} else {
										result = Color.ORANGE;
									}
								}
							}
						} catch (Exception ex) {
							getLog().error(ex);
							result = Color.GRAY;
						}
						return result;
					}
				});

		
		KeyListener keyListener = new KeyListener() {
			
			@Override
			public void keyTyped(KeyEvent e) {
				if ('o' == e.getKeyChar()){
					//display operations
					StringBuilder result = new StringBuilder();
					logOperatorsAndDetails(result);
					getLog().info(result);
				}
				
			}
			
			@Override
			public void keyReleased(KeyEvent e) {
			}
			
			@Override
			public void keyPressed(KeyEvent e) {
			}
		};
		displayConfiguration.getKeyListeners().add(keyListener);
		
		return displayConfiguration;
	}

	/**
	 * Appends the ordered operators and their details to a text
	 * 
	 * @param result
	 *            the text
	 */
	public void logOperatorsAndDetails(StringBuilder result) {

		boolean displayGlobalPreconditions = false;
		/*
		 * =JOptionPane.YES_OPTION == JOptionPane .showConfirmDialog(null,
		 * "display tables?");
		 */
		for (BoostedNode node : getGraph().getAllVerticesOrdered()) {
			Operation op = getOperation(node);

			// display global precondition
			Object globalPrecondition = op.getGlobalPrecondition();
			if (globalPrecondition != null
					&& globalPrecondition instanceof Table) {
				if (displayGlobalPreconditions) {
					result.append(((Table) globalPrecondition).getGuiString())
						.append("\n");
				}
			}

			// display operator
			result.append("\n").append(op.getOperator()).append(" ")
					.append(op.getOperatorDetails());
		}

	}

	@Override
	public void warning(TransformerException exception)
			throws TransformerException {
		warningHappened = true;
	}

	@Override
	public void error(TransformerException exception)
			throws TransformerException {
		errorHappened = true;
	}

	@Override
	public void fatalError(TransformerException exception)
			throws TransformerException {
		fatalErrorHappened = true;
	}

	/**
	 * Tests the integrity of this plan
	 */
	public void selfcheck() throws Exception {

		for (BoostedNode node : graph.getAllVertices()) {
			getOperation(node).checkOperator();

			Refactor.checkNoDuplicateEdges(node);

		}

		for (BoostedNode node : graph.getAllSourceVertices()) {
			Operation operation = getOperation(node);
			Check.notNull(operation.getGlobalPrecondition(),
					"globalPrecondition of " + operation.toString());
		}

	}
}
