package net.sf.gluebooster.demos.pojo.planning;

import java.util.Collection;
import java.util.Hashtable;

import javax.naming.spi.ObjectFactory;

import net.sf.gluebooster.demos.pojo.refactor.BoostedNodeGraph;
import net.sf.gluebooster.demos.pojo.refactor.Condition;
import net.sf.gluebooster.demos.pojo.refactor.GraphDisplayConfiguration;
import net.sf.gluebooster.demos.pojo.refactor.Operation;
import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.basic.container.BoostedNodeUtils;
import net.sf.gluebooster.java.booster.essentials.meta.HasName;
import net.sf.gluebooster.java.booster.essentials.objects.BoostedObject;

import org.apache.commons.collections15.Transformer;

/**
 * Configuration of (one step of) the planning.
 *
 */
public class Planning<Conditio extends Condition<Conditio, NameOfObjects>, NameOfObjects>
		extends
		BoostedObject {

	/**
	 * The generated plan.
	 */
	private Plan<Conditio, NameOfObjects> plan;

	/**
	 * The environment of the planning.
	 */
	private PlannerEnv env;

	/**
	 * The configuration of the default display
	 */
	private GraphDisplayConfiguration<BoostedNode, NameOfObjects> defaultDisplayConfiguration;

	public Planning() {

	}

	public Planning(Plan<Conditio, NameOfObjects> plan, PlannerEnv env) {
		this.plan = plan;
		this.env = env;
	}

	/**
	 * Create a planning with a plan with one node
	 * 
	 * @param planWithOnlyOneOperation
	 *            the operation of the node
	 * @param world
	 *            the world of the planning
	 */
	public Planning(
			Operation<Conditio, NameOfObjects> planWithOnlyOneOperation,
			World world) throws Exception {
		plan = new Plan<Conditio, NameOfObjects>();
		plan.initialize(world, planWithOnlyOneOperation);
		setEnv(new PlannerEnv());
	}

	public Plan<Conditio, NameOfObjects> getPlan() {
		return plan;
	}

	public void setPlan(Plan<Conditio, NameOfObjects> plan) {
		this.plan = plan;
	}

	public PlannerEnv getEnv() {
		return env;
	}

	public void setEnv(PlannerEnv env) {
		this.env = env;
	}

	/**
	 * Log some information at info level
	 * 
	 * @param objectsToLog
	 *            the information
	 */
	public void info(Object... objectsToLog) {
		getEnv().getPlanningLog().info(objectsToLog);
	}

	/**
	 * Log some information at debug level
	 * 
	 * @param objectsToLog
	 *            the information
	 */
	public void debug(Object... objectsToLog) {
		getEnv().getPlanningLog().info(objectsToLog);
	}

	/**
	 * Log some information at warn level
	 * 
	 * @param objectsToLog
	 *            the information
	 */
	public void warn(Object... objectsToLog) {
		getEnv().getPlanningLog().info(objectsToLog);
	}

	/**
	 * Log some information at error level
	 * 
	 * @param objectsToLog
	 *            the information
	 */
	public void error(Object... objectsToLog) {
		getEnv().getPlanningLog().info(objectsToLog);
	}

	/**
	 * Transforms an unsolved node of the plan into a subgraph and replaces the node by the subgraph.
	 * 
	 * @param unsolvedNodeOfPlan
	 *            the node to be transformed
	 * @param transformer
	 *            the transformer of the node
	 * @return true if something has been done
	 */
	public boolean transformNode(BoostedNode unsolvedNodeOfPlan,
			Transformer<BoostedNode, BoostedNodeGraph> transformer)
			throws Exception {
		
		plan.selfcheck();
		BoostedNodeGraph subgraph;
		if (transformer instanceof ObjectFactory){
			// if the plan can be passed
			subgraph = (BoostedNodeGraph) ((ObjectFactory) transformer)
					.getObjectInstance(unsolvedNodeOfPlan, null, null,
							createHashtableWithPlan(plan));
			plan.selfcheck();
		} else {
			subgraph = transformer.transform(unsolvedNodeOfPlan);
		}

		if (subgraph == null) {
			return false; // nothing has been done
		} else {
			// Check operations
			for (BoostedNode node : subgraph.getAllVertices()) {
				plan.getOperation(node).checkOperator();
			}

			// add preconditions of node to sources
			Operation<Conditio, NameOfObjects> unsolvedPlanNode = Operation
					.getOperation(unsolvedNodeOfPlan);
			Collection<BoostedNode> newSourceVertices = subgraph
					.getAllSourceVertices();
			if (newSourceVertices.isEmpty()) {
				// use the sucessors instead
				newSourceVertices = BoostedNodeUtils
						.getSuccessors(unsolvedNodeOfPlan);
			}

			plan.addPreconditionToNodes(newSourceVertices,
					unsolvedPlanNode.getPrecondition());
			plan.addGlobalPreconditionToNodes(newSourceVertices,
					unsolvedPlanNode.getGlobalPrecondition());

			plan.addDefaultConditionsOfOperators(subgraph);

			plan.replaceNode(unsolvedNodeOfPlan, subgraph);

			plan.selfcheck();
			return true;
		}

	}

	/**
	 * Instantiates a virtual object with a real object.
	 * 
	 * @param nodeWithVirtualObject
	 *            the node which virtual object is to be replaced
	 * @param virtualObject
	 *            the object to replace
	 * @param concreteObject
	 *            the replacement
	 */
	public void instantiateVirtualObject(BoostedNode nodeWithVirtualObject,
			WorldObject<NameOfObjects, ?> virtualObject,
			HasName<NameOfObjects> concreteObject)
			throws Exception {

		getPlan().instantiateVirtualObject(plan.getGraph(),
				nodeWithVirtualObject,
				virtualObject,
				concreteObject);

	}

	/**
	 * Create a hashtable that contains a plan. This method must be coordinated with getPlanFromHashtable.
	 * 
	 * @param plan
	 *            the given plan
	 * @return the key of the plan is Plan.class
	 */
	public static Hashtable<?, ?> createHashtableWithPlan(Plan plan) {
		Hashtable environmentWithPlan = new Hashtable();
		environmentWithPlan.put(Plan.class, plan);
		return environmentWithPlan;
	}

	/**
	 * Extracts a plan from the hashtable. This method must be coordinated with createHashtableWithPlan.
	 * 
	 * @param env
	 *            must contain the plan with key Plan.class
	 * @return the found plan.
	 */
	public static <Conditio extends Condition<Conditio, NameOfObjects>, NameOfObjects> Plan<Conditio, NameOfObjects> getPlanFromHashtable(
			Hashtable env) {
		return (Plan<Conditio, NameOfObjects>) env.get(Plan.class);
	}

	/**
	 * Remove superfluous operator nodes from the plan.
	 * 
	 * @return true if something has changed
	 */
	public boolean removeSuperfluousOperators() throws Exception {
		return getPlan().removeSuperfluousOperators();
	}

	public GraphDisplayConfiguration<BoostedNode, NameOfObjects> getDefaultDisplayConfiguration() {
		return defaultDisplayConfiguration;
	}

	public void setDefaultDisplayConfiguration(GraphDisplayConfiguration<BoostedNode, NameOfObjects> defaultDisplayConfiguration) {
		this.defaultDisplayConfiguration = defaultDisplayConfiguration;
	}

	/**
	 * Does a simple display of the plan.
	 */
	public void defaultDisplay() throws Exception {
		if (defaultDisplayConfiguration != null) {
			getPlan().simpleDisplay(defaultDisplayConfiguration);
		}

	}
}
