package net.sf.gluebooster.demos.pojo.planning;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.naming.spi.ObjectFactory;
import javax.xml.transform.ErrorListener;
import javax.xml.transform.TransformerException;

import net.sf.gluebooster.demos.pojo.refactor.BoostedNodeGraph;
import net.sf.gluebooster.demos.pojo.refactor.Condition;
import net.sf.gluebooster.demos.pojo.refactor.Operation;
import net.sf.gluebooster.demos.pojo.refactor.Operator;
import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.essentials.objects.BoostedObject;

import org.apache.commons.collections15.Transformer;
import org.apache.commons.lang3.tuple.Pair;

/**
 * Information about the word in which the planning takes place.
 * 
 * @author CBauer
 * @defaultParamText node the node to be inspected/modified
 *
 */
public abstract class World<Conditio extends Condition<Conditio, NameOfObjects>, NameOfObjects>
		extends
		BoostedObject {

	// protected static final String OPERATORNAME_NOTHING_TO_DO =
	// "NOTHING TODO DO";

	/**
	 * Name of the operator that states that a node/problem is still to solve.
	 */
	protected static final String OPERATORNAME_TO_SOLVE = "TO SOLVE";

	/**
	 * Name of the operator that states that a new virtual object is to be
	 * created.
	 */
	protected static final String OPERATORNAME_NEW_VIRTUAL_OBJECT = "NEW VIRTUAL OBJECT";

	/**
	 * Name of the operator that states that virtual object is to be deleted.
	 */
	protected static final String OPERATORNAME_DELETE_VIRTUAL_OBJECT = "DELETE VIRTUAL OBJECT";

	/**
	 * The names of properties that may be conflicting.
	 * 
	 * Example: In a block world a property may be "move block" to denote that
	 * an operator moves a block. Another property may be "new non free block",
	 * to denote that a block is covered by another.
	 * 
	 * If in parallel branches the same block is moved, then there may be a
	 * possible conflict. So the pair would be ("move block"/"move block").
	 * 
	 * If a block is covered in one branch and moved in a parallel branch there
	 * may be a conflict, too. So another pair would be
	 * ("move block"/"new non free block").
	 */
	private Collection<Pair<String, String>> possiblePropertyConflictsDefinition;

	/**
	 * Listeners that listen to the computation of global postconditions.
	 */
	private List<ErrorListener> computeGlobalPostconditionListener = new ArrayList<ErrorListener>();

	/**
	 * 
	 * @param possiblePropertyConflicts
	 *            the value of possiblePropertyConflictsDefinition
	 */
	protected World(Collection<Pair<String, String>> possiblePropertyConflicts) {
		this.possiblePropertyConflictsDefinition = possiblePropertyConflicts;
	}

	// protected static final String OPERATORNAME_UNDEFINED = "UNDEFINED";

	public Collection<Pair<String, String>> getPossiblePropertyConflictsDefinition() {
		return possiblePropertyConflictsDefinition;
	}

	/**
	 * Get all operators of the world
	 * 
	 * @return all operators
	 */
	public abstract Collection<Operator<NameOfObjects>> getAllOperators();

	/**
	 * Create a TO_SOLVE-Operator
	 * 
	 * @return the created operator
	 */
	public final Operator<NameOfObjects> getToSolveOperator() {
		return getOperator(OPERATORNAME_TO_SOLVE);
	}

	/**
	 * Get a specific operator
	 * 
	 * @param operatorname
	 *            the name of the operator
	 * @return the found operator
	 * @throws IllegalStateException
	 *             if no operator is found
	 */
	public final Operator<NameOfObjects> getOperator(String operatorname) {
		for (Operator<NameOfObjects> operator : getAllOperators()) {
			if (operatorname.equals(operator.getName()))
				return operator;
		}
		throw new IllegalStateException("No operator found for " + operatorname);

	}

	/**
	 * Is it necessary to do something to solve a given node.
	 * 
	 * @return true if nothing needs to be done with the node
	 */
	public abstract boolean isNoNeedToSolve(Operation node) throws Exception;

	/**
	 * Has the node a to-solve-operator
	 * 
	 * @return true if the operator is to-solve
	 */
	public boolean isToSolveOperator(Operation node) {
		return getToSolveOperator().isOperatorOf(node);
	}

	/**
	 * Computes the maximum abstraction level of all operators.
	 * 
	 * @return the computed level
	 */
	public int getMaximumOperatorAbstractionLevel() {
		int result = getToSolveOperator().getAbstractionLevel();
		for (Operator operator : getAllOperators()) {
			result = Math.max(result, operator.getAbstractionLevel());
		}
		return result;
	}

	/**
	 * Computes the minimum abstraction level of all operators.
	 * 
	 * @return the computed level
	 * 
	 */
	public int getMinimumOperatorAbstractionLevel() {
		int result = getToSolveOperator().getAbstractionLevel();
		for (Operator operator : getAllOperators()) {
			result = Math.min(result, operator.getAbstractionLevel());
		}
		return result;
	}

	/**
	 * Return the solving algorithms that are capable to solve an unsolved node.
	 * If the solving algorithms can not do anything they return null; otherwise
	 * they always have to return a node of a tree. The root of the tree may be
	 * a
	 * 
	 * 
	 * Default expander are implemented here.
	 * 
	 * @return Collection of ObjectFactory with signature BoostedNodeGraph
	 *         getObjectInstance(Operation unsolvedNode, ...)
	 */
	public final Collection<Transformer<BoostedNode, BoostedNodeGraph>> getExpander(
			BoostedNode node) {
		ArrayList<Transformer<BoostedNode, BoostedNodeGraph>> result = new ArrayList<Transformer<BoostedNode, BoostedNodeGraph>>(
				getWorldSpecificExpander(node));

		if (isOperator(node, OPERATORNAME_TO_SOLVE) && result.isEmpty()) {
			// split the postcondition into single conditions if no other
			// expander is found
			getLog().warn(
					"World.getExpander splitting of postconditions is not implemented");

			// new_postconditions = postcondition->split();
			// new_postconditions->reset();
			// while ( new_postconditions->next(&next_postcondition))
			// {
			// next_netnode_object = ( PLAN_OBJECT *)(this->deepcopy());
			// next_netnode_object->deepdelete_postcondition();
			// next_netnode_object->postcondition =( CONDITION *)
			// next_postcondition;
			// new_nodes.insert(next_netnode_object);
			// }
			// DELETE(new_postconditions);
			// net->insert_before( node_nr, new_nodes);
			// net->remove(node_nr);
			// result = TRUE ;

		}

		return result;
	}

	/**
	 * Determines factories that can initialize a virtual object
	 * 
	 * @param nodeWithVirtualObject
	 *            node containing a virtual object
	 * @return ObjectFactory with Pair<virtualObject, concreteObject> getObjectInstance(Object nodeWithVirtualObject, Name ignored, Context ignoredCtx,
	 *         Hashtable<?, ?> environmentWithPlan must contain the plan under key Plan.class
	 */
	public abstract Set<ObjectFactory> getVirtualObjectInitializer(
			BoostedNode nodeWithVirtualObject);

	/**
	 * Return the world-specific solving algorithms that are capable to solve an
	 * unsolved node.
	 * 
	 * The solving algorithms need to create global pre- and postconditions.
	 * This is done in the MetaPlanner
	 * 
	 * This method should be overwritten by subclasses
	 * 
	 * @return the possible solving algorithms
	 */
	public Collection<Transformer<BoostedNode, BoostedNodeGraph>> getWorldSpecificExpander(
			BoostedNode node) {
		return Collections.emptyList();
	}

	/**
	 * Has the node a given operator
	 * 
	 * @param operatorname
	 *            the name of the operator
	 * @return true if the operators are equal
	 */
	public boolean isOperator(BoostedNode node, String operatorname) {
		return extractOperation(node).hasOperatorname(operatorname);
	}

	/**
	 * Extracts the operation of a node.
	 * 
	 * @return the operation of the node
	 */
	protected Operation<Conditio, NameOfObjects> extractOperation(
			BoostedNode node) {
		return Operation.getOperation(node);
	}

	/**
	 * Gets the properties of a node.
	 * 
	 * @return A mapping of the names of the properties to the names of the objects that have the property. Not all properties to be in this map.
	 * 
	 */
	public abstract Map<String, Set<String>> getPropertyValues(
			Operation<Conditio, NameOfObjects> node);

	/**
	 * Gets the names of the objects that have a given property
	 * 
	 * @param nameOfProperty
	 *            the name of the property
	 * @return not null
	 */
	public Set<String> getObjectsWithProperty(
			Operation<Conditio, NameOfObjects> node,
			String nameOfProperty) {
		Set<String> result;
		Map<String, Set<String>> propertyValues = getPropertyValues(node);
		if (propertyValues == null) {
			result = new HashSet<String>();
		} else {
			result = propertyValues.get(nameOfProperty);

			if (result == null)
				result = new HashSet<String>();
		}
		return result;
	}

	/**
	 * Has an operator of a node a virtual property May be overwritten by subclasses for a better implementation.
	 * 
	 * @return true, if it has
	 */
	public boolean hasVirtualProperty(Operation<Conditio, NameOfObjects> node) {
		return OPERATORNAME_NEW_VIRTUAL_OBJECT.equals(node.getOperator()
				.getName());
	}

	/**
	 * Has a node a virtual property.
	 * 
	 * @return true, if it has
	 */
	public boolean hasVirtualProperty(BoostedNode node) {
		return hasVirtualProperty(extractOperation(node));
	}

	/**
	 * Get the default postcondition for the operation of a node
	 * 
	 * @param plan
	 *            the plan of the node
	 * @return null if no default postcondition exists
	 */
	public abstract Conditio getDefaultPostcondition(
			Plan<Conditio, NameOfObjects> plan,
			Operation<Conditio, NameOfObjects> node) throws Exception;

	/**
	 * Get the default precondition for the operation of a node
	 * 
	 * @param plan
	 *            the plan of the node
	 * @return null if no default precondition exists
	 */
	public abstract Conditio getDefaultPrecondition(
			Plan<Conditio, NameOfObjects> plan,
			Operation<Conditio, NameOfObjects> node) throws Exception;

	/**
	 * Get the default deletecondition for the operation of a node
	 * 
	 * @param plan
	 *            the plan of the node
	 * @return null if no default deletecondition exists
	 */
	public abstract Conditio getDefaultDeletecondition(
			Plan<Conditio, NameOfObjects> plan,
			Operation<Conditio, NameOfObjects> node) throws Exception;

	/**
	 * If an operation has a global precondition then try to compute its global postcondition
	 * 
	 * @param operation
	 *            the inspected operation
	 */
	public final void updateGlobalPostcondition(
			Operation<Conditio, NameOfObjects> operation)
			throws Exception {
		operation.setGlobalPostcondition(null);// reset
		Conditio globalPrecondition = operation.getGlobalPrecondition();
		if (globalPrecondition != null) {
			operation.setGlobalPostcondition(computeGlobalPostcondition(
					operation, globalPrecondition));
		}
		
		// BLOCKWORLD_PLAN_OBJECT::no_need_to_solve()
	}

	/**
	 * Compute the global postcondition for an operation
	 * 
	 * @param globalPrecondition
	 *            the global precondition for which the postcondition will be computed
	 * @param operation
	 *            the operation is applied to the global precondition
	 * @return null if nothing can be computed
	 */
	protected abstract Conditio computeGlobalPostcondition(
			Operation<Conditio, NameOfObjects> operation,
			Conditio globalPrecondition)
			throws Exception;

	/**
	 * Computes a global postcondition by adding/removing the postcondition/deletecondition
	 * 
	 * @param operation
	 *            contains the post- and delete-condition
	 * @param globalPrecondition
	 *            the starting point of the computation
	 * @return global precondition + postcondition - delete condition
	 */
	protected Conditio computeGlobalPostconditionByCondtions(
			Operation<Conditio, NameOfObjects> operation,
			Conditio globalPrecondition)
			throws Exception {
		Conditio globalPostcondition = globalPrecondition.cloneMe();

		Conditio deleteCondition = operation.getDeleteCondition();
		if (deleteCondition != null) {
			globalPostcondition.remove(deleteCondition);
		}

		Conditio postcondition = operation.getPostcondition();
		if (postcondition != null) {
			globalPostcondition.add(postcondition);
		}

		return globalPostcondition;
	}


	public List<ErrorListener> getComputeGlobalPostconditionListener() {
		return computeGlobalPostconditionListener;
	}

	public void setComputeGlobalPostconditionListener(
			List<ErrorListener> computeGlobalPostconditionListener) {
		this.computeGlobalPostconditionListener = computeGlobalPostconditionListener;
	}

	/**
	 * Warns the global postcondition listener.
	 * 
	 * @param objectsToLog
	 *            the warning
	 */
	protected void globalPostconditionWarning(Object... objectsToLog) {
		getLog().warn(objectsToLog);
		String message = "";
		for (int i = 0; i < objectsToLog.length; i++) {
			message += (objectsToLog[i]);
		}
		TransformerException exception = new TransformerException(message);
		for (ErrorListener listener : getComputeGlobalPostconditionListener()) {
			try {
				listener.warning(exception);
			} catch (Exception ex) {
				// nothing to do
			}
		}
	}

	/**
	 * Create a new operation
	 * 
	 * @param globalPrecondition
	 *            global precondition of the new operation
	 * @param operator
	 *            operator of the new operation
	 * @param postcondition
	 *            postcondition of the new operation
	 * @param solvingPriority
	 *            solving priority of the new operation
	 * @return the created operation
	 */
	public Operation<Conditio, NameOfObjects> createOperation(
			Conditio globalPrecondition, Operator<NameOfObjects> operator,
			Conditio postcondition, int solvingPriority) {
		Operation<Conditio, NameOfObjects> result = createOperation();
		result.setGlobalPrecondition(globalPrecondition);
		result.setPostcondition(postcondition);
		result.setOperator(getToSolveOperator());
		result.setSolvingPriority(Integer.MAX_VALUE);
		return result;

	}

	/**
	 * Create a default operation. Must be overwritten if a subclass of Operation is wanted.
	 * 
	 * @return a new operation (without content)
	 */
	protected Operation<Conditio, NameOfObjects> createOperation() {
		return new Operation<Conditio, NameOfObjects>();
	}

}
