package net.sf.gluebooster.demos.pojo.refactor;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import net.sf.gluebooster.java.booster.essentials.utils.ThrowableBoostUtils;

/**
 * Abstract implemeentation of the view logic of an application (without display).
 * 
 * @author cbauer
 *
 * @param <Controller>
 */
public abstract class AppViewAbstraction<Controller extends CommandHandler, Definition extends AppDefinition> extends BoostedObject
		implements Runnable, MouseListener, ActionListener, CommandHandler<Object, Object, Object> {

	/**
	 * The initial state of the view.
	 */
	protected static final int STATE_INITIAL = 0;

	/**
	 * The current state of this view.
	 */
	protected int state = STATE_INITIAL;

	/**
	 * The controller of this view (model-view-controller).
	 */
	private CommandHandler<Object, Object, Map<Object, Object>> controller;

	/**
	 * The model used to display in the view. (Not necessarily the model used by the controller)
	 */
	private Map<Object, Object> displayModel;

	/**
	 * The title of the view/application.
	 */
	private String title;

	/**
	 * The definition of the application.
	 */
	private Definition appDefinition;

	/**
	 * Initialization and start of the application.
	 * 
	 * @param view
	 *            the view to start
	 */
	@SuppressWarnings("rawtypes")
	public static <Controller extends CommandHandler> void startApplication(AppViewAbstraction<Controller, ?> view,
			CommandHandler controller) {
		view.setController(controller);
		view.run();
		// TODO use command instead of run()
	}

	// public AppViewAbstraction(String title) {
	// this.title = title;
	// }

	/**
	 * @param definition
	 *            the application definition
	 */
	public AppViewAbstraction(Definition definition) {
		this.title = definition.getTitle();
		setAppDefinition(definition);
	}


	public CommandHandler<Object, Object, Map<Object, Object>> getController() {
		return controller;
	}

	public void setController(CommandHandler controller) {
		this.controller = controller;
	}

	protected int getState() {
		return state;
	}

	protected void setState(int state) {
		this.state = state;
	}

	@Override
	public final void run() {

		try {
			// TODO do i have to do this?
			displayModel.put(AppDefinition.FIELD_TITLE, title);

			((CommandHandler) displayModel.get(AppDefinition.COMMAND_HANDLER)).executeCommand(AppDefinition.COMMAND_START, null);
			executeCommand(AppDefinition.COMMAND_START, null);
		} catch (Exception ex) {
			throw ThrowableBoostUtils.toRuntimeException(ex);
		}
	}

	@Override
	public void mousePressed(MouseEvent e) {

	}

	@Override
	public void mouseReleased(MouseEvent e) {

	}

	@Override
	public void mouseEntered(MouseEvent e) {

	}

	@Override
	public void mouseExited(MouseEvent e) {

	}

	@Override
	public void mouseClicked(MouseEvent e) {
	}

	/**
	 * Executes the command of the event. If subclasses want handle special cases they must overwrite this method
	 */
	@Override
	public void actionPerformed(ActionEvent e) {
		try {
			executeCommand(e.getActionCommand(), null);
		} catch (Exception ex) {
			ex.printStackTrace();
			// TODO display feedback
		}

	}

	public Map<Object,Object> getDisplayModel() {
		return displayModel;
	}

	public void setDisplayModel(Map<Object, Object> display) {
		this.displayModel = display;
	}

	/**
	 * Initializes and starts an application
	 * 
	 * @param controllerClass
	 *            the class of the controller
	 * @param controllerModel
	 *            the model of the controller
	 * @param viewClass
	 *            the class of the view
	 * @param displayClass
	 *            the class of the display
	 * @return the created view
	 */
	public static AppViewAbstraction startApplication(Class<? extends AppControllerAbstraction> controllerClass, Map controllerModel,
			Class<? extends AppViewAbstraction> viewClass,
			Class displayClass) throws Exception {
		AppControllerAbstraction controller = controllerClass.newInstance();
		if (controllerModel != null) {
			controller.setModel(controllerModel);
		}

		AppViewAbstraction view = viewClass.newInstance();

		displayClass.getConstructor(AppViewAbstraction.class).newInstance(view);
		startApplication(view, controller);
		return view;

	}

	/**
	 * Starts a user interaction that displays a file dialog
	 * 
	 * @param dialog
	 *            the configuration of the dialog.
	 * @return the chosen file.
	 */
	protected File getFileFromUser(DialogConfiguration dialog) throws Exception {
		File result = ((UserInteraction) getDisplayModel().get(AppDefinition.USER_INTERACTION)).showFileDialog(dialog);
		return result;
	}

	/**
	 * Execute a command by the controller.
	 * 
	 * @param command
	 *            the command to be invoked
	 * @param parameter
	 *            parameter of the command
	 * @return the result of the command
	 */
	protected Map<Object, Object> controllerCommand(Object command, Object... parameter) throws Exception {
		return getController().executeCommand(command, parameter);
	}

	/**
	 * Display a file dialog, then let the controller execute a command with the chosen file as parameter.
	 * 
	 * @param command
	 *            the command to be executed
	 * @param fileDialog
	 *            the configuration of the dialog
	 * @return the result of the command.
	 */
	protected Map<Object, Object> controllerCommandWithFile(Object command, DialogConfiguration fileDialog) throws Exception {
		fileDialog.setThrowExceptionIfNothingSelected(false);
		File file = getFileFromUser(fileDialog);
		if (file != null) {
			return controllerCommand(command, file);
		}
		// TODO feedback
		return null;
	}

	/**
	 * Copies some entries from the controller model to the display model
	 * 
	 * @param keys
	 *            the keys of the entries
	 */
	protected void copyFromControllerToDisplayModel(Collection<Object> keys) throws Exception {
		Map<Object, Object> controllerModel = getController().executeCommand(AppDefinition.COMMAND_GET_MODEL, null);
		ContainerBoostUtils.copy(controllerModel, displayModel, keys, true);
	}

	/**
	 * Copies some entries from the controller model to the display model
	 * 
	 * @param keys
	 *            the keys of the entries
	 */
	protected void copyFromControllerToDisplayModel(Object... keys) throws Exception {
		Map<Object, Object> controllerModel = getController().executeCommand(AppDefinition.COMMAND_GET_MODEL, null);
		ContainerBoostUtils.copyVararg(controllerModel, displayModel, true, keys);
	}


	/**
	 * Lets the controller execute a command with some fields as parameters (map)
	 * 
	 * @param command
	 *            the command to be executed
	 * @param fieldnames
	 *            the names of the fields to be transferred (as map) to the controller
	 * @return the result of the command
	 */
	protected Map<Object, Object> controllerCommandWithFields(Object command, Collection<Object> fieldnames) throws Exception {
		Map<Object, Object> selectedFields = new HashMap<Object, Object>();
		ContainerBoostUtils.copy(displayModel, selectedFields, fieldnames, true);
		return controllerCommand(command, selectedFields);

	}

	public Definition getAppDefinition() {
		return appDefinition;
	}

	public void setAppDefinition(Definition appDefinition) {
		this.appDefinition = appDefinition;
	}

}
