package net.sf.gluebooster.demos.pojo.refactor;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.basic.container.BoostedNodeUtils;
import net.sf.gluebooster.java.booster.basic.container.SimpleBoostedNode;
import net.sf.gluebooster.java.booster.essentials.objects.BoostedObject;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

/**
 * A graph of BoostedNodes.
 * 
 * @see http://en.wikipedia.org/wiki/Glossary_of_graph_theory
 * @see http://primes.utm.edu/graph/
 * @author CBauer
 *
 * @defaultParamText xTranslation translation along the x-axis
 * @defaultParamText displayConfiguration configuration of the display
 *
 */
public class BoostedNodeGraph extends BoostedObject {

	/**
	 * Nodes of the connected components of the graph. For each connected
	 * component at least one node must be in this collection. The graph
	 * contains all nodes that are reachable from these nodes.
	 * 
	 */
	private BoostedHashSetExtension<BoostedNode> someNodesOfTheConnectedComponents;

	/**
	 * Should there be at most 1 edge between nodes.
	 */
	private boolean atMost1EdgeBetweenNodes = false;

	public BoostedNodeGraph() {
		someNodesOfTheConnectedComponents = new BoostedHashSetExtension<BoostedNode>();
		someNodesOfTheConnectedComponents.setThrowExceptionWhenEmptied(true);
	}

	/**
	 * Creates a graph with existing nodes.
	 * 
	 * @param nodeOfConnectedGraph
	 *            one node of the graph
	 */
	public BoostedNodeGraph(BoostedNode nodeOfConnectedGraph) {
		this();
		someNodesOfTheConnectedComponents.add(nodeOfConnectedGraph);
	}

	public BoostedHashSetExtension<BoostedNode> getSomeNodesOfTheConnectedComponents() {
		return someNodesOfTheConnectedComponents;
	}

	public void setSomeNodesOfTheConnectedComponents(
			Set<BoostedNode> someNodesOfTheConnectedComponents) {
		this.someNodesOfTheConnectedComponents.clear();
		this.someNodesOfTheConnectedComponents
				.addAll(someNodesOfTheConnectedComponents);
	}

	/**
	 * A sink vertex in a directed graph is a vertex without outgoing edge. See http://en.wikipedia.org/wiki/Vertex_%28graph_theory%29
	 * 
	 * @return all sink vertices in a new set
	 */
	public Set<BoostedNode> getAllSinkVertices() {
		
		HashSet<BoostedNode> result = new HashSet<BoostedNode>();
		for (BoostedNode node : someNodesOfTheConnectedComponents) {
			result.addAll(Refactor.getAllSinkVertices(node));
		}

		return result;
	}

	/**
	 * A source vertex in a directed graph is a vertex without ingoing edge. See http://en.wikipedia.org/wiki/Vertex_%28graph_theory%29
	 * 
	 * @return all source vertices in a new set
	 */
	public Set<BoostedNode> getAllSourceVertices() {

		HashSet<BoostedNode> result = new HashSet<BoostedNode>();
		for (BoostedNode node : someNodesOfTheConnectedComponents) {
			result.addAll(Refactor.getAllSourceVertices(node));
		}

		return result;
	}

	/**
	 * Gets all vertices of this graph
	 * 
	 * @return all vertices in a new set
	 */
	public Set<BoostedNode> getAllVertices() {
		HashSet<BoostedNode> result = new HashSet<BoostedNode>();
		for (BoostedNode node : someNodesOfTheConnectedComponents) {
			result.addAll(BoostedNodeUtils.getAllNodesOfGraph(node));
		}
		return result;
	}

	/**
	 * Adds a node to the graph
	 * 
	 * @param node
	 *            the node to be added
	 * @return the node
	 */
	public BoostedNode addNode(BoostedNode node) {
		someNodesOfTheConnectedComponents.add(node);
		return node;
	}

	/**
	 * Adds a new node with a given value
	 * 
	 * @param value
	 *            the value of the new node
	 * @return the created node
	 */
	public BoostedNode addNodeWithValue(Object value) {
		SimpleBoostedNode node = new SimpleBoostedNode();
		node.getAttributes().setValue(value);
		addNode(node);
		return node;
	}

	/**
	 * Replaces the node by a subgraph. The nodes of the subgraph are changed.
	 * 
	 * @param node
	 *            the node to be replaced
	 * @param replaceBy
	 *            the replacing subgraph
	 */
	public void replace(BoostedNode node, BoostedNodeGraph replaceBy)
			throws Exception {

		// checkConnections();

		// Cache the predecessors and successors before they may be changed
		// during the following algorithm
		Collection<BoostedNode> nodePredecessors = BoostedNodeUtils
				.getPredecessors(node);
		Collection<BoostedNode> nodeSuccessors = BoostedNodeUtils
				.getSuccessors(node);

		// Begin of display
		// ArrayList<BoostedNode> outInterested = new ArrayList<BoostedNode>(
		// nodePredecessors);
		// outInterested.add(node);
		// ArrayList<BoostedNode> inInterested = new ArrayList<BoostedNode>(
		// nodeSuccessors);
		// inInterested.add(node);
		//
		// if (replaceBy == null) {
		// new BoostedNodeVisualization(outInterested, inInterested);
		// }
		// End of display


		Refactor.replace(node, replaceBy);

		// if (replaceBy == null) {
		// new BoostedNodeVisualization(outInterested, inInterested);
		// }

		// add the new nodes before removing the old one, so that the graph is
		// never without nodes (and throws an exception)
		if (replaceBy == null || replaceBy.isEmpty()) {
			someNodesOfTheConnectedComponents.addAll(nodePredecessors);
			someNodesOfTheConnectedComponents.addAll(nodeSuccessors);
		} else {
			someNodesOfTheConnectedComponents.addAll(replaceBy
				.getSomeNodesOfTheConnectedComponents());
		}

		someNodesOfTheConnectedComponents.remove(node);

		removeRedundancies();

		// checkConnections();

	}

	/**
	 * Adds a new edge
	 * 
	 * @param from
	 *            start of the edge
	 * @param to
	 *            end of the edge
	 */
	public void addEdge(BoostedNode from, BoostedNode to) throws Exception {
		Refactor.addEdge(from, to);

	}

	/**
	 * Test method to check whether the connections between the nodes are
	 * correctly set
	 */
	public void checkConnections() throws Exception {
		for (BoostedNode node : getAllVertices()) {
			Refactor.checkPredecessorSuccessor(node);
		}
	}

	/**
	 * Removes the node. Edges to/from the node are redirected to the successors/predecessors.
	 * 
	 * @param node
	 *            the node to be removed
	 */
	public void removeNode(BoostedNode node) throws Exception {
		// getLog().warn("refactor removeNode");
		// simpleDisplay();
		replace(node, null);
	}

	/**
	 * Orders the vertices according to their pre- and successors
	 * 
	 * @return the ordered vertices
	 */
	public List<BoostedNode> getAllVerticesOrdered() {

		ArrayList<BoostedNode> result = new ArrayList<BoostedNode>();
		HashSet<BoostedNode> toInspect = new HashSet<BoostedNode>(
				getAllSourceVertices());


		while (!toInspect.isEmpty()) {
			boolean changed = false;
			HashSet<BoostedNode> nextToInspect = new HashSet<BoostedNode>();
			HashSet<BoostedNode> inspected = new HashSet<BoostedNode>();
			for (BoostedNode node : toInspect) {
				if (result.contains(node)) {
					inspected.add(node);
					changed = true;
				} else {
					if (result.containsAll(BoostedNodeUtils
							.getPredecessors(node))) {
						result.add(node);
						inspected.add(node);
						nextToInspect.addAll(BoostedNodeUtils
								.getSuccessors(node));
						changed = true;
					}
				}
			}
			toInspect.removeAll(inspected);
			toInspect.addAll(nextToInspect);

			if (!changed) {
				simpleDisplay();

				StringBuilder analysis = new StringBuilder(
						"Could not order. There seems to be a loop in the graph\n");
				if (result.isEmpty()) {
					analysis.append("There is no source node in the graph");
				} else {
					analysis.append("nodes that are ok: ").append(result)
							.append("\n\n");
					HashSet<BoostedNode> handled = new HashSet<BoostedNode>();
					for (BoostedNode node : result) {
						for (BoostedNode successor : BoostedNodeUtils
								.getSuccessors(node)) {
							if (!handled.contains(successor)) {
								handled.add(successor);
								HashSet<BoostedNode> predecessors = new HashSet<BoostedNode>(
										BoostedNodeUtils
												.getPredecessors(successor));
								predecessors.removeAll(result);
								analysis.append("node "
										+ successor
										+ " \n has the following predecessors that prevent its ordering:\n"
										+ predecessors);
							}

						}
					}
				}

				throw new IllegalStateException(analysis.toString());
			}
		}


		return result;

	}
	
	/**
	 * Precondition: this graph must be a forest
	 * 
	 * @param toThisNodes
	 *            the nodes to be inspected
	 * @return for each node the parallel node
	 */
	public Set<Pair<BoostedNode, BoostedNode>> getParallelNodesOfForest(
			Collection<BoostedNode> toThisNodes) {
		Set<Pair<BoostedNode, BoostedNode>> result = new HashSet<Pair<BoostedNode, BoostedNode>>();
		// getLog().error("getParallelNodes is not yet implemented");
		Set<BoostedNode> allVertices = getAllVertices();

		for (BoostedNode node : toThisNodes) {
			HashSet<BoostedNode> parallelNodes = new HashSet<BoostedNode>(
					allVertices);
			parallelNodes.removeAll(getConnectedComponentNodes(node));
			for (BoostedNode parallelNode : parallelNodes) {
				result.add(new ImmutablePair<BoostedNode, BoostedNode>(node,
						parallelNode));
			}
		}

		return result;
	}

	/**
	 * Finds connected components
	 * 
	 * @param containingNode
	 *            one node of the component
	 * @return the nodes of the connected component.
	 */
	public Set<BoostedNode> getConnectedComponentNodes(
			BoostedNode containingNode) {
		HashSet<BoostedNode> result = new HashSet<BoostedNode>();
		result.add(containingNode);
		Refactor.addReachableSuccessors(containingNode, result);
		Refactor.addReachablePredecessors(containingNode, result);
		return result;
	}
	

	/**
	 * Are there any vertices
	 * 
	 * @return false if no vertices present
	 */
	public boolean isEmpty() {
		return getAllVertices().isEmpty();
	}

	/**
	 * Are two nodes connected
	 * 
	 * @param node1
	 *            the first node
	 * @param node2
	 *            the second node
	 * @return true if connected
	 */
	public boolean areConnected(BoostedNode node1, BoostedNode node2) {
		return getConnectedComponentNodes(node1).contains(node2)
				|| getConnectedComponentNodes(node2).contains(node1);
	}

	/**
	 * Gets the sequence between two nodes. The two nodes must be connected by a sequence
	 * 
	 * @param startnode
	 *            the beginning of the sequence
	 * @param endnode
	 *            the end of the sequence
	 * @return the sequence
	 */
	public List<BoostedNode> getSequenceBetween(BoostedNode startnode,
			BoostedNode endnode) {
		ArrayList<BoostedNode> result = new ArrayList<BoostedNode>();
		BoostedNode node = startnode;
		do {
			result.add(node);
			node = BoostedNodeUtils.getSuccessors(node).get(0);
		} while (!node.equals(endnode));

		return result;
	}

	/**
	 * Displays the graph (in a simple layout).
	 */
	public void simpleDisplay() {
		GraphDisplayConfiguration<BoostedNode, String> displayConfiguration = GraphDisplayConfiguration
				.createFilledNodesWithoutTextConfiguration();
		displayConfiguration.setDisplayVertexValueInsteadOfName(true);
		displayConfiguration.setVertexDetailDisplayLevel(10);

		simpleDisplay(displayConfiguration);
	}

	/**
	 * Displays the graph.
	 * 
	 * @param displayConfiguration
	 *            configuration of the display
	 */
	public void simpleDisplay(
			GraphDisplayConfiguration<BoostedNode, ?> displayConfiguration) {
		simpleDisplay(displayConfiguration, new Flag(true, false));
	}

	/**
	 * Displays the graph
	 * 
	 * @param displayConfiguration
	 * @param overlappingAllowed
	 *            if false an exception will occur if there is overlapping (of nodes)
	 */
	public void simpleDisplay(
			GraphDisplayConfiguration<BoostedNode, ?> displayConfiguration,
			Flag overlappingAllowed) {

		Refactor.createFrame(displayConfiguration.getTitle(), 800, 800, true,
				Refactor.createOverviewGraphComponent(this,
						displayConfiguration, overlappingAllowed));

	}

	public boolean isAtMost1EdgeBetweenNodes() {
		return atMost1EdgeBetweenNodes;
	}

	public void setAtMost1EdgeBetweenNodes(boolean atMost1EdgeBetweenNodes) {
		this.atMost1EdgeBetweenNodes = atMost1EdgeBetweenNodes;
	}

	/**
	 * Removes duplicate edges (if only at most 1 edge is allowed)
	 */
	public void removeRedundancies() {
		if (atMost1EdgeBetweenNodes) {
			for (BoostedNode vertex: getAllVertices()){
				Refactor.removeDuplicateEdges(vertex);
			}
			
		}
	}

	/**
	 * Returns a map that represents the structure of the graph. The keys are all vertices of the graph. If a vertex has no successor, its successorlist is
	 * empty. This map can be used to store/analyse the structure of the graph
	 * 
	 * @return the created map
	 */
	public Map<Object, List<Object>> getSuccessorStructureMap() {
		HashMap<Object, List<Object>> result = new HashMap<Object, List<Object>>();
		for (BoostedNode node : getAllVertices()) {
			ArrayList<Object> successors = new ArrayList<Object>();
			result.put(node.getAttributes().getId(), successors);
			for (BoostedNode successor : BoostedNodeUtils.getSuccessors(node)) {
				successors.add(successor.getAttributes().getId());
			}
		}

		return result;
	}

	/**
	 * Creates a graph from a successor map
	 * 
	 * @param successorMap
	 *            infos about the edges
	 * @return the created graph
	 */
	public static BoostedNodeGraph createFromSuccessorStructureMap(
			Map<Object, List<Object>> successorMap) throws Exception {
		int counter = 0;
		BoostedNodeGraph result = new BoostedNodeGraph();
		Map<Object, BoostedNode> idNodeMapping = new HashMap<Object, BoostedNode>();
		for (Object nodeid : successorMap.keySet()) {
			BoostedNode node = new SimpleBoostedNode<Object>();
			node.getAttributes().setId("" + counter++);
			idNodeMapping.put(nodeid, node);
			result.addNode(node);
		}

		for (Map.Entry<Object, List<Object>> entry : successorMap.entrySet()) {
			BoostedNode node = idNodeMapping.get(entry.getKey());
			for (Object successorId : entry.getValue()) {
				BoostedNode successor = idNodeMapping.get(successorId);
				result.addEdge(node, successor);
			}
		}

		return result;
	}
	
	/**
	 * Is a walk between two nodes possible
	 * 
	 * @param from
	 *            the start of the walk
	 * @param to
	 *            the end of the walk
	 * @return true if possible
	 */
	public boolean isWalk(BoostedNode from, BoostedNode to) {
		HashSet<BoostedNode> successors = new HashSet<BoostedNode>();
		Refactor.addReachableSuccessors(from, successors);
		return successors.contains(to);
	}
}
