package net.sf.gluebooster.demos.pojo.refactor;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

import javax.swing.JFrame;

import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.basic.container.SimpleBoostedNode;
import net.sf.gluebooster.java.booster.basic.gui.swing.SwingFactory;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

import edu.uci.ics.jung.algorithms.layout.FRLayout;
import edu.uci.ics.jung.algorithms.layout.Layout;
import edu.uci.ics.jung.graph.DirectedSparseMultigraph;
import edu.uci.ics.jung.visualization.VisualizationViewer;

/**
 * Graphical display of one BoostedNode, its In- and Out-Points and the
 * connection to the pre- and successors. It has methods to vizalize the
 * changing of the connections.
 * 
 * @author CBauer
 *
 */
public class BoostedNodeVisualization {

	/**
	 * The graph of the points
	 */
	private DirectedSparseMultigraph<BoostedNode, String> graph = new DirectedSparseMultigraph();

	/**
	 * The labels of the points
	 */
	private HashMap<BoostedNode, String> labels = new HashMap<BoostedNode, String>();

	/**
	 * The edges between the points
	 */
	private Set<Pair<BoostedNode, BoostedNode>> edges = new HashSet<Pair<BoostedNode, BoostedNode>>();
	
	/**
	 * A counter to name the edges.
	 */
	private int counter = 0;

	/**
	 * Constructor with values.
	 * 
	 * @param nodesWithInterestInOutgoingEdges
	 *            of which node is interest in the outgoing edges.
	 * @param nodesWithInterestInIngoingEdges
	 *            of which node is interest in the ingoing edges.
	 */
	public BoostedNodeVisualization(
			Collection<BoostedNode> nodesWithInterestInOutgoingEdges,
			Collection<BoostedNode> nodesWithInterestInIngoingEdges) {


		


		for (BoostedNode node : nodesWithInterestInOutgoingEdges) {
			addNodeIfNecessary(node, ""
					+ ((SimpleBoostedNode) node).getAttributes().getValue());

			for (BoostedNode outpoint : node.getOutPoints()) {
				addEdge(node, outpoint);
				addEdges(outpoint, node.getSuccessorInPoints(outpoint));

			}

		}

		for (BoostedNode node : nodesWithInterestInIngoingEdges) {
			addNodeIfNecessary(node, ""
					+ ((SimpleBoostedNode) node).getAttributes().getValue());

			for (BoostedNode inpoint : node.getInPoints()) {
				addEdge(inpoint, node);
				addEdges(node.getPredecessorOutPoints(inpoint), inpoint);
			}

		}

		Layout<BoostedNode, String> layout = new FRLayout<BoostedNode, String>(
				graph);
		layout.setSize(new Dimension(300, 300));
		VisualizationViewer<BoostedNode, String> visualization = new VisualizationViewer<BoostedNode, String>(
				layout);


		visualization.setPreferredSize(new Dimension(700, 700));
		// visualization.getRenderContext().setVertexFillPaintTransformer(
		// new TransformerByMap<BoostedNode, Paint>(nodePaint));
		// visualization
		// .setVertexToolTipTransformer(new TransformerByMap<BoostedNode,
		// String>(
		// labels));

		visualization.getRenderContext().setVertexLabelTransformer(TransformerByMap.createGetFromMapTransformer(labels));

		int size = 800;
		SwingFactory guiFactory = new SwingFactory();
		JFrame frame = guiFactory.createFrame("BoostedNodeVisualization", size,
				size, true, new BorderLayout());

		frame.getContentPane().add(visualization, BorderLayout.CENTER);

		frame.setVisible(true);

	}

	/**
	 * If the node is not yet present, add it.
	 * 
	 * @param node
	 *            the new node
	 * @param label
	 *            the label of the node
	 */
	private void addNodeIfNecessary(BoostedNode node, String label) {
		if (!labels.containsKey(node)) {
			labels.put(node, label);
			graph.addVertex(node);
		}

	}

	/**
	 * Adds an edge between two nodes
	 * 
	 * @param from
	 *            the starting of the edge
	 * @param to
	 *            the end of the edge
	 */
	private void addEdge(BoostedNode from, BoostedNode to){
		addNodeIfNecessary(from, "");
		addNodeIfNecessary(to, "");

		Pair<BoostedNode, BoostedNode> edge = new ImmutablePair<BoostedNode, BoostedNode>(
				from, to);
		if (!edges.contains(edge)) {
			edges.add(edge);
			graph.addEdge("" + counter++, from, to);
		}
	}

	/**
	 * Adds edges from one node to multiple nodes
	 * 
	 * @param from
	 *            the start of the edges
	 * @param to
	 *            the ends of the edges
	 */
	private void addEdges(BoostedNode from, Collection<BoostedNode> to) {
		for (BoostedNode node : to) {
			addEdge(from, node);
		}
	}

	/**
	 * Adds edges from multiple nodes to one node
	 * 
	 * @param from
	 *            the start of the edges
	 * @param to
	 *            the end of the edges
	 */
	private void addEdges(Collection<BoostedNode> from, BoostedNode to) {
		for (BoostedNode node : from) {
			addEdge(node, to);
		}
	}

}
