package net.sf.gluebooster.demos.pojo.refactor;

import java.util.Set;

import net.sf.gluebooster.java.booster.essentials.meta.HasName;

/**
 * Conditions (predicates) for planning.
 * 
 * @author CBauer
 * @defaultParamText otherCondition The other condition
 * 
 */
public interface Condition<Implementation extends Condition<Implementation, NameOfObjects>, NameOfObjects>
		extends Cloneable {

	/**
	 * Does this condition imply another condition
	 * 
	 * @return true if this => otherCondition
	 * @throws Exception
	 */
	Boolean implies(Condition otherCondition) throws Exception;

	/**
	 * Is this condition in conflict with another condition Example: If this
	 * condition states that the color of an object is red and another condition
	 * states that it is white, they are in conflict.
	 * 
	 * @return true if a conflict exists
	 */
	Boolean isInConflict(Condition otherCondition) throws Exception;

	// void copyFrom(Condition condition);
	// replace by new and add

	/**
	 * Is this condition an elemenarty condition. Elementary conditions are not composed from other conditions.
	 * 
	 * @return true, if elementary
	 */
	boolean isElementary() throws Exception;

	// returns TRUE if the condition isn´t a conjunction of subconditions

	/**
	 * Splits this condition into (elementary) conditions. Only AND-Conditions
	 * are split.
	 * 
	 * @return the subconditions that can be composed with AND to this condition
	 */
	<Result extends Condition> Set<Result> split() throws Exception;

	// returns subconditions; the conjunction of the subconditions is the
	// condition

	/**
	 * Subtracts another condition from this condition. Example if this
	 * condition states that an object is black and round, another condition
	 * states that the object is black then the difference is that the object is
	 * black.
	 * 
	 * @return this - otherCondition
	 */
	Condition difference(Implementation otherCondition) throws Exception;

	// subtracts the condition from this condition

	/**
	 * Removes the parts of another condition from this condition. Example if
	 * this condition states that an object is black and round, another
	 * condition states that the object is black then this condition after the
	 * removal is that the object is black.
	 */
	void remove(Condition otherCondition) throws Exception;

	// subtracts the condition from this

	/**
	 * Adds the parts of another condition from this condition. Example if this
	 * condition states that an object is round, another condition states that
	 * the object is black then this condition after the add is that the object
	 * is black and round.
	 */
	void add(Condition otherCondition) throws Exception;
	// adds a deepcopy of the condition to this condition

	/**
	 * Create a new, empty Implementation instance. The instance is prepared
	 * (general values are already set).
	 * 
	 * @return the created condition
	 */
	Condition createEmptyCondition() throws Exception;

	/**
	 * If a given object is found in this condition, its name is replaced by a new name.
	 * 
	 * @param object
	 *            the object to be searched for
	 * @param newName
	 *            the new name of the object
	 */
	void replaceObjectName(HasName<NameOfObjects> object,
			NameOfObjects newName) throws Exception;

	/**
	 * Clones this instance.
	 * 
	 * @return the clone
	 */
	Implementation cloneMe() throws Exception;

	/**
	 * Tries to find a subcondition that is like the given template.
	 * 
	 * @param template
	 *            the template of the subcondition
	 * @return null if nothing is found
	 */
	<Result extends Condition> Result findSubcondition(Result template)
			throws Exception;
}
