package net.sf.gluebooster.demos.pojo.refactor;

import java.awt.Color;
import java.awt.Paint;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import net.sf.gluebooster.java.booster.basic.container.SimpleBoostedNode;
import net.sf.gluebooster.java.booster.essentials.meta.HasName;
import net.sf.gluebooster.java.booster.essentials.utils.ThrowableBoostUtils;

import org.apache.commons.collections15.Transformer;

import edu.uci.ics.jung.visualization.control.GraphMouseListener;
import edu.uci.ics.jung.visualization.renderers.Renderer;

/**
 * Configuration how a graph is to be displayed
 * 
 * @author CBauer
 *
 */
public class GraphDisplayConfiguration<Vertex, Edge> {

	/**
	 * The title of this configuration.
	 */
	private String title = "Graph";

	/**
	 * Should the vertex labels be displayed.
	 */
	private boolean displayVertexLabel = false;

	/**
	 * The position of the label relative to the center of the vertex.
	 */
	private Renderer.VertexLabel.Position vertexLabelPosition = Renderer.VertexLabel.Position.CNTR;

	/**
	 * Generates the label for a vertex
	 */
	private Transformer<Vertex, String> vertexLabelTransformer;

	/**
	 * Should the tooltips of vertices be displayed.
	 */
	private boolean displayVertexTooltip = true;

	/**
	 * Generates the tooltip for a vertex.
	 */
	private Transformer<Vertex, String> vertexToolTipTransformer;

	/**
	 * Should the label of a vertex be the value or the name (default).
	 */
	private boolean displayVertexValueInsteadOfName = false;

	/**
	 * The level how detailed a vertex should be displayed.
	 */
	private int vertexDetailDisplayLevel;

	/**
	 * The vertices that should be highlighted.
	 */
	private Collection<Vertex> highlightedVertices;

	/**
	 * The colors of the vertices
	 */
	private Map<Vertex, Color> vertexColors;

	/**
	 * Listeners for user keyboard interaction
	 */
	private List<KeyListener> keyListeners = new ArrayList<KeyListener>();

	/**
	 * Generates the shape of a vertex.
	 */
	private Transformer<Vertex, ? extends Shape> vertexShapeTransformer;


	/**
	 * The background color of vertices.
	 */
	private Transformer<Vertex, Paint> vertexBackgroundTransformer;

	/**
	 * Listener for user mouse interaction.
	 */
	private GraphMouseListener<Vertex> clickListener;

	/**
	 * Creates a listener that logs the details of the clicked vertex
	 * 
	 * @return the created listener
	 */
	public GraphMouseListener<Vertex> createLoggingVertexDetailsListener() {
		Transformer<Vertex, String> vertexMaxDetailStringer = createToStringTransformer(
				isDisplayVertexValueInsteadOfName(), Integer.MAX_VALUE);
		GraphMouseListener<Vertex> result = createLoggingVertexClickListener(vertexMaxDetailStringer);
		return result;
	}


	/**
	 * The background color of vertices.
	 * 
	 * @return the created transformer
	 */
	public Transformer<Vertex, Paint> createDefaultNodesBackgroundTransformer() {
		return new Transformer<Vertex, Paint>() {

			@Override
			public Paint transform(Vertex node) {
				if (highlightedVertices != null
						&& highlightedVertices.contains(node)) {
					return Color.GREEN;
				} else if (vertexColors != null
						&& vertexColors.containsKey(node)) {
					return vertexColors.get(node);
				} else {
					return Color.BLUE;
				}
			}
		};
	}

	/**
	 * Returns a transformer that creates a rectangle accoring the size of the label of the node.
	 * 
	 * @return the created transformer
	 */
	public Transformer<Vertex, Rectangle> createRectangleForLabelTransformer() {
		return new Transformer<Vertex, Rectangle>() {

			@Override
			public Rectangle transform(Vertex node) {
				return new Rectangle(vertexLabelTransformer.transform(node)
						.length() * 7, 20); // TODO adapt to size
			}
		};
	}

	/**
	 * Create a transformer that extracts information of a vertex.
	 * 
	 * @param displayValueInsteadOfName
	 *            should the value be displayed instead of the name
	 * @param displayLevel
	 *            the level of details
	 * @return the created transformer
	 */
	public static <VertexOrEdge> Transformer<VertexOrEdge, String> createToStringTransformer(
			final boolean displayValueInsteadOfName, final int displayLevel) {

		return new Transformer<VertexOrEdge, String>() {
			@Override
			public String transform(VertexOrEdge vertex) {
				try {
					// modified here
					// TODO do some refactoring
					// SimpleBoostedNode should inherit from BoostedObject
					// BoostedObject should get the objectAttributes and should
					// be
					// stringable
					// There should be an interface HasAttributes and the
					// objectattributes should be stringable

					Object objectToDisplay = vertex;// default

					if (displayValueInsteadOfName) {
						if (vertex instanceof SimpleBoostedNode) {
							objectToDisplay = ((SimpleBoostedNode) vertex)
									.getAttributes().getValue();
						}
					}

					if (objectToDisplay == null) {
						return "null";
					} else if (objectToDisplay instanceof Stringable) {
						return ((Stringable) objectToDisplay)
								.toString(displayLevel);
					} else if (objectToDisplay instanceof HasName) {
						return "" + ((HasName) objectToDisplay).getName();
					} else {
						return objectToDisplay.toString();
					}

					/*
					 * String name = ""; if (node instanceof HasName) name =
					 * name + ((HasName) node).getName(); else name = name +
					 * node; return name;
					 */
				} catch (Exception ex) {
					throw ThrowableBoostUtils.toRuntimeException(ex);
				}
			}
		};
	}

	/**
	 * Creates a transformer that changes vertices to strings.
	 * 
	 * @return the created transformer
	 */
	public Transformer<Vertex, String> createDefaultVertexToStringTransformer() {
		return createToStringTransformer(isDisplayVertexValueInsteadOfName(),
				getVertexDetailDisplayLevel());
	}

	/**
	 * Create a configuration for a graph with rectangular vertices and text as label.
	 * 
	 * @return the created configuration
	 */
	public static GraphDisplayConfiguration createRectangularNodesWithTextConfiguration() {
		GraphDisplayConfiguration result = new GraphDisplayConfiguration();

		result.setDisplayVertexLabel(true);
		// displayConfiguration.setVertexBackgroundTransformer(null);
		result.setVertexShapeTransformer(result
				.createRectangleForLabelTransformer());
		result.setVertexLabelTransformer(result
				.createDefaultVertexToStringTransformer());

		result.setClickListener(result.createLoggingVertexDetailsListener());
		return result;
	}
	

	/**
	 * Creates the configuration of a graph with colored nodes without labels but tooltips.
	 * 
	 * @return the created configuration
	 */
	public static GraphDisplayConfiguration createFilledNodesWithoutTextConfiguration() {
		GraphDisplayConfiguration result = new GraphDisplayConfiguration();
		result.vertexBackgroundTransformer = result
				.createDefaultNodesBackgroundTransformer();
		result.setVertexToolTipTransformer(result
				.createDefaultVertexToStringTransformer());

		result.setClickListener(result.createLoggingVertexDetailsListener());
		return result;
	}
	
	/**
	 * Creates a configuration of a graph with simple display and tooltips.
	 * 
	 * @return the created configuration
	 */
	public static GraphDisplayConfiguration createUnadornedConfiguration() {
		GraphDisplayConfiguration result = new GraphDisplayConfiguration();
		result.setVertexToolTipTransformer(result
				.createDefaultVertexToStringTransformer());

		result.setClickListener(result.createLoggingVertexDetailsListener());
		return result;
	}

	/**
	 * Private so that the create methods need to be used
	 */
	private GraphDisplayConfiguration() {

	}

	public Transformer<Vertex, Paint> getVertexBackgroundTransformer() {
		return vertexBackgroundTransformer;
	}

	public void setVertexBackgroundTransformer(
			Transformer<Vertex, Paint> vertexBackgroundTransformer) {
		this.vertexBackgroundTransformer = vertexBackgroundTransformer;
	}

	public boolean isDisplayVertexValueInsteadOfName() {
		return displayVertexValueInsteadOfName;
	}

	public void setDisplayVertexValueInsteadOfName(
			boolean displayVertexValueInsteadOfName) {
		this.displayVertexValueInsteadOfName = displayVertexValueInsteadOfName;
	}

	public int getVertexDetailDisplayLevel() {
		return vertexDetailDisplayLevel;
	}

	public void setVertexDetailDisplayLevel(int vertexDetailDisplayLevel) {
		this.vertexDetailDisplayLevel = vertexDetailDisplayLevel;
	}

	public String getTitle() {
		return title;
	}

	public void setTitle(String title) {
		this.title = title;
	}

	public Collection<Vertex> getHighlightedVertices() {
		return highlightedVertices;
	}

	public void setHighlightedVertices(Collection<Vertex> highlightedVertices) {
		this.highlightedVertices = highlightedVertices;
	}

	public Map<Vertex, Color> getVertexColors() {
		return vertexColors;
	}

	public void setVertexColors(Map<Vertex, Color> vertexColors) {
		this.vertexColors = vertexColors;
	}

	public List<KeyListener> getKeyListeners() {
		return keyListeners;
	}

	public Transformer<Vertex, ? extends Shape> getVertexShapeTransformer() {
		return vertexShapeTransformer;
	}

	public void setVertexShapeTransformer(
			Transformer<Vertex, ? extends Shape> vertexShapeTransformer) {
		this.vertexShapeTransformer = vertexShapeTransformer;
	}

	public boolean isDisplayVertexLabel() {
		return displayVertexLabel;
	}

	public void setDisplayVertexLabel(boolean displayVertexLabel) {
		this.displayVertexLabel = displayVertexLabel;
	}

	public boolean isDisplayVertexTooltip() {
		return displayVertexTooltip;
	}

	public void setDisplayVertexTooltip(boolean displayVertexTooltip) {
		this.displayVertexTooltip = displayVertexTooltip;
	}

	public Renderer.VertexLabel.Position getVertexLabelPosition() {
		return vertexLabelPosition;
	}

	public void setVertexLabelPosition(
			Renderer.VertexLabel.Position vertexLabelPosition) {
		this.vertexLabelPosition = vertexLabelPosition;
	}

	public Transformer<Vertex, String> getVertexLabelTransformer() {
		return vertexLabelTransformer;
	}

	public void setVertexLabelTransformer(
			Transformer<Vertex, String> vertexLabelTransformer) {
		this.vertexLabelTransformer = vertexLabelTransformer;
	}

	public Transformer<Vertex, String> getVertexToolTipTransformer() {
		return vertexToolTipTransformer;
	}

	public void setVertexToolTipTransformer(
			Transformer<Vertex, String> vertexToolTipTransformer) {
		this.vertexToolTipTransformer = vertexToolTipTransformer;
	}

	/**
	 * Creates a listener that logs the nodes that are clicked.
	 * 
	 * @param logTextCreator
	 *            creates the text from the vertices
	 * @return the created listener
	 */
	public static <Vertex> GraphMouseListener<Vertex> createLoggingVertexClickListener(
			final Transformer<Vertex, String> logTextCreator) {

		GraphMouseListener<Vertex> clickListener = new GraphMouseListener<Vertex>() {

			@Override
			public void graphClicked(Vertex v, MouseEvent me) {
				String text = logTextCreator.transform(v);
				System.out.println(text);

			}

			@Override
			public void graphPressed(Vertex v, MouseEvent me) {

			}

			@Override
			public void graphReleased(Vertex v, MouseEvent me) {

			}
		};

		return clickListener;
	}

	public void setKeyListeners(List<KeyListener> keyListeners) {
		this.keyListeners = keyListeners;
	}

	public GraphMouseListener<Vertex> getClickListener() {
		return clickListener;
	}

	public void setClickListener(GraphMouseListener<Vertex> clickListener) {
		this.clickListener = clickListener;
	}

}
