package net.sf.gluebooster.demos.pojo.refactor;


import java.awt.Color;
import java.awt.Paint;
import java.util.Hashtable;
import java.util.List;
import java.util.Set;

import javax.naming.Context;
import javax.naming.Name;
import javax.naming.spi.ObjectFactory;

import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.basic.container.BoostedNodeUtils;
import net.sf.gluebooster.java.booster.basic.meta.DocumentationContext;
import net.sf.gluebooster.java.booster.basic.meta.HasDocumentation;
import net.sf.gluebooster.java.booster.basic.text.general.BookBoostUtils;
import net.sf.gluebooster.java.booster.basic.text.general.DocumentationResourcesId;
import net.sf.gluebooster.java.booster.essentials.meta.HasName;
import net.sf.gluebooster.java.booster.essentials.objects.BoostedObject;
import net.sf.gluebooster.java.booster.essentials.utils.ThrowableBoostUtils;

import org.apache.commons.collections15.Transformer;
import org.apache.commons.lang.ObjectUtils;

/**
 * duplicate of the
 * net.sf.gluebooster.java.booster.basic.math.TransformationGraph Replace the
 * old implementation by this implementation.
 * 
 * {@Feature}A graph of transformations. Multiple transformations can
 * be done sequentially or parallel. {@FeatureEnd}
 * 
 * The grahp must have exactly one root (node with in-degree 0) and one leaf
 * with out-degree 0.
 * 
 * The graph is traversed to compute the result.
 * 
 * The nodes (vertices) must be TransformationNodes. The attributes of the
 * boosted nodes must contain an Attribute with key ObjectFactory.class which
 * getObjectInstance is executed during the traversal. The result of the
 * getObjectInstance is used as object of the next invocation.
 * 
 * 
 * If multiple inputs are to be converted the root node of the graph must
 * transform a map (keys are the name of the in-nodes, values the values of the
 * in-nodes).
 * 
 * @author CBauer
 * 
 * @defaultParamText node the inspected node
 * @defaultParamText transformation the transformation that the node will
 *                   execute.
 * @defaultParamText inOrOutPoint the inspected point of a node
 * @defaultParamText nodeOrPoint the inspected node or point
 * 
 */
public class TransformationGraph extends BoostedObject implements
		ObjectFactory, HasDocumentation, CommandHandler {

	/**
	 * The nodes must be TransformationNodes.
	 */
	private BoostedNodeGraph graph = new BoostedNodeGraph();

	/**
	 * Should the transformation of a node be executed even if no
	 * in-point-values have been set. TODO: better explanation example
	 * electrical circuits versus one-time-computation
	 */
	private boolean computeTransformationsEvenIfInPointsHaveNotBeenSet = false;
	/**
	 * Is it a continuous transformation, so that each step each node has to be
	 * computed (like an electric current).
	 */
	private boolean continuousTransformation = false;

	/**
	 * Should the graph be displayed even if errors occur.
	 */
	private boolean displayTheGraphWhenAnErrorOccurs = false;


	/**
	 * Gets the root node of this graph. If the graph is no tree, one of the
	 * roots is returned.
	 * 
	 * @return The first source node
	 */
	public TransformationNode getRootNode() {
		return (TransformationNode) graph.getAllSourceVertices().iterator()
				.next();
	}

	/**
	 * Gets the result of the graph. If the graph has multiple result nodes, one
	 * of them is returned.
	 * 
	 * @return the first sink
	 */
	public TransformationNode<?> getResultNode() {
		return (TransformationNode<?>) graph.getAllSinkVertices().iterator()
				.next();
	}

	/**
	 * Sets the value of the first in-point of the root node.
	 * 
	 * @param value
	 *            the new value
	 */
	public void setCurrentValueOfOfRootNodeInPoint(Object value) {
		BoostedNode inPoint = (BoostedNode) getRootNode().getInPoints().get(0);
		TransformationNode.setCurrentValue(inPoint, value);
	}

	// public void setCurrentValueSetOfOfRootNodeInPoint(boolean flag) {
	// BoostedNode inPoint = getRootNode().getInPoints().get(0);
	// TransformationNode.setCurrentValueSet(inPoint, flag);
	// }


	/**
	 * Create a new node with a transformation.
	 * 
	 * @param transformation
	 *            the transformation the node will perform
	 * @param transformationNeedsObjectToTransform
	 *            does the transformation need an input object
	 * @return the created node
	 */
	public BoostedNode addNode(ObjectFactory transformation,
			boolean transformationNeedsObjectToTransform) throws Exception {
		TransformationNode result = new TransformationNode(transformation);
		graph.addNode(result);
		return result;
	}

	/**
	 * Only one transformation can take place after the other. TODO use the
	 * environment to implement a max of the steps if there is continuous
	 * computation.
	 */
	@Override
	public synchronized Object getObjectInstance(Object value, Name name,
			Context nameCtx, Hashtable<?, ?> environment) throws Exception {

		if (computeTransformationsEvenIfInPointsHaveNotBeenSet) {
			throw new IllegalStateException(
					"can not compute result, because the transformation will not end. computeTransformationsEvenIfInPointsHaveNotBeenSet must be set to false");

		}

		checkAndResetGraph(true, true, true);
		setRootOfGraphValue(value);

		while (transformStep()) {
			// repeat as long as something changes
		}

		return getResultNode().getCurrentValue();

	}


	/**
	 * Checks whether this graph is configured sufficiently. Values may be
	 * changed.
	 * 
	 * @param removeOutPointValues
	 *            should the out-point values be cleared
	 * @param removeInPointValues
	 *            should the in-point values be cleared
	 * @param setStateNotHandled
	 *            should the state of the node be set to 'not yet handled'
	 */
	public void checkAndResetGraph(boolean removeOutPointValues,
			boolean removeInPointValues, boolean setStateNotHandled) {

		int sourceVerticesCount = graph.getAllSourceVertices().size();
		if (sourceVerticesCount != 1)
			throw new IllegalStateException(
					"no singular root of graph defined: " + sourceVerticesCount);

		int sinkVerticesCount = graph.getAllSinkVertices().size();
		if (sinkVerticesCount != 1)
			throw new IllegalStateException("no singular sink node defined: " + sinkVerticesCount);

		for (BoostedNode node : graph.getAllVertices()) {

			TransformationNode<?> trafoNode = (TransformationNode) node;

			if (trafoNode.getTransformation() == null)
				throw new IllegalStateException("no transformation in node "
						+ trafoNode);

			if (removeOutPointValues) {
				trafoNode.removeOutpointValues();
			}

			if (removeInPointValues) {
				trafoNode.removeInpointValues();
			}

			if (setStateNotHandled) {
				trafoNode.setStateNotYetHandled();
			}
		}

	}

	public void setRootOfGraphValue(Object value) {
		List<BoostedNode> inPoints = getRootNode().getInPoints();
		if (inPoints.isEmpty() && value != null) {
			throw new IllegalStateException(
					"no in-point, but a value to transform");
		} else if (inPoints.size() > 1) {
			throw new IllegalStateException(
					"only at most 1 in-point is supported");
		} else if (inPoints.size() == 1) {
			TransformationNode.setCurrentValue(inPoints.get(0), value);
		}

	}


	/**
	 * Executes a transformation of the in-node values. If
	 * computeTransformationsEvenIfInPointsHaveNotBeenSet is true, then the
	 * transformation is done by all nodes. Otherwise it is only done by the
	 * nodes which in-node-values have been set.
	 * 
	 * @return true, if something has been done
	 * @throws Exception
	 */
	public boolean transformStep() throws Exception {

		try {
			boolean result = false;

			checkAndResetGraph(continuousTransformation, false,
					continuousTransformation);

			result |= computeOutPointValues();
			// getLog().debug("transformStep part result: ", result);
			result |= transferValuesFromOutPointsToInPoints();

			// getLog().debug("transformStep result: ", result);
			return result;
		} catch (Exception ex) {
			if (displayTheGraphWhenAnErrorOccurs) {
				simpleDisplay();
			}
			throw ex;
		}
	}

	/**
	 * compute the new out-point values.TODO make private later
	 * 
	 * @return true, if at least one new value has been computed.
	 * @throws Exception
	 */
	public boolean computeOutPointValues() throws Exception {
		boolean result = false;
		for (BoostedNode transformationNode : graph.getAllVertices()) {
			TransformationNode<?> node = (TransformationNode) transformationNode;
				if (continuousTransformation || node.isStateNotYetHandled()) {
				result |= node
						.transform(computeTransformationsEvenIfInPointsHaveNotBeenSet);
				}
		}

		return result;
	}

	/**
	 * transfer the values of preceeding out-points to the in-points. TODO make
	 * private
	 * 
	 * @return true, if at least one new value has been computed.
	 */
	public boolean transferValuesFromOutPointsToInPoints() {

		boolean result = false;

		for (BoostedNode node : graph.getAllVertices()) {

			for (BoostedNode inPoint : node.getInPoints()) {
				List<BoostedNode> predecessorOutPoints = node
						.getPredecessorOutPoints(inPoint);

				if (predecessorOutPoints.size() == 0
						&& getRootNode().equals(node)) {
					// the root node may have no predecessors out points
					// nothing needs to be done
				} else {

					if (predecessorOutPoints.size() != 1) {
						throw ThrowableBoostUtils
								.createNotImplementedException("multiple or no predecessor out-points are not yet supported");
					}
					BoostedNode predecessorOutPoint = predecessorOutPoints
							.get(0);
					if (TransformationNode.hasCurrentValue(predecessorOutPoint)) {
						Object outPointValue = TransformationNode
								.getCurrentValue(predecessorOutPoint);

						// check whether something will change
						if (TransformationNode.hasCurrentValue(inPoint)) {
							Object inPointValue = TransformationNode
									.getCurrentValue(inPoint);
							if (!ObjectUtils
									.equals(outPointValue, inPointValue)) {
								result = true;
								// getLog().debug(
								// "set result to true, because in-point-value changed");
							}
						}

						TransformationNode.setCurrentValue(inPoint,
								outPointValue);

					} else if (computeTransformationsEvenIfInPointsHaveNotBeenSet) {
						throw ThrowableBoostUtils
								.createNotImplementedException("transfer not yet implemented");

					}
				}
			}
		}
		return result;
	}

	/**
	 * Get a short text with the name and state of an in-point, out-point or
	 * vertex
	 * 
	 * @return the state of the nodeOrPoint
	 */
	public String getShortState(BoostedNode nodeOrPoint) {
		Set<BoostedNode> nodes = BoostedNodeUtils
				.getAllNodesOfGraph(getRootNode());
		for (BoostedNode node : nodes) {

			if (node.equals(nodeOrPoint))
				return "" + getName(nodeOrPoint);

			for (BoostedNode inPoint : node.getInPoints()) {
				if (nodeOrPoint.equals(inPoint))
					return getShortInPointState(nodeOrPoint);
			}

			for (BoostedNode point : node.getOutPoints()) {
				if (nodeOrPoint.equals(point))
					return getShortOutPointState(nodeOrPoint);
			}

		}

		throw new IllegalStateException("No element of graph: " + nodeOrPoint);
	}

	/**
	 * Gets the name of a node or point.
	 * 
	 * @return The name if the nodeOrPoint is a HasName, otherwise that
	 *         nodeOrPoint
	 */
	private static Object getName(BoostedNode nodeOrPoint) {
		if (nodeOrPoint instanceof HasName)
			return ((HasName) nodeOrPoint).getName();
		else
			return nodeOrPoint;

	}

	/**
	 * Get a short text with the name and state of an in-point
	 * 
	 * @param inPoint
	 *            the inspected point
	 * @return describing text of the inPoint
	 */
	public String getShortInPointState(BoostedNode inPoint) {
		StringBuilder name = new StringBuilder();

		name.append(getName(inPoint));

		name.append("(");
		if (TransformationNode.isCurrentValueSet(inPoint))
			name.append("!");

		name.append(" ");
		if (TransformationNode.hasCurrentValue(inPoint))
			name.append(TransformationNode.getCurrentValue(inPoint));
		name.append(" )");

		return name.toString();

	}

	/**
	 * Get a short text with the name and state of an out-point
	 * 
	 * @param outPoint
	 *            the inspected point
	 * @return describing text of the inPoint
	 */
	public String getShortOutPointState(BoostedNode outPoint) {
		// do (at the moment) the same as with in-points
		return getShortInPointState(outPoint);
	}

	// public BoostedNode getResultNode() {
	// return resultNode;
	// }
	//
	// public void setResultNode(BoostedNode resultNode) {
	// this.resultNode = resultNode;
	// }

	// private Set<BoostedNode> getAllNodesOfGraph() {
	// if (unmodifiableGraph && allNodesOfGraph != null)
	// return allNodesOfGraph;
	// Set<BoostedNode> result = BoostedNodeUtils
	// .getAllNodesOfGraph(getRootNode());
	// if (unmodifiableGraph)
	// allNodesOfGraph = result;
	//
	// return result;
	// }

	// public boolean isUnmodifiableGraph() {
	// return unmodifiableGraph;
	// }
	//
	// public void setUnmodifiableGraph(boolean unmodifiableGraph) {
	// this.unmodifiableGraph = unmodifiableGraph;
	// }

	public boolean isComputeTransformationsEvenIfInPointsHaveNotBeenSet() {
		return computeTransformationsEvenIfInPointsHaveNotBeenSet;
	}

	public void setComputeTransformationsEvenIfInPointsHaveNotBeenSet(
			boolean computeTransformationsEvenIfInPointsHaveNotBeenSet) {
		this.computeTransformationsEvenIfInPointsHaveNotBeenSet = computeTransformationsEvenIfInPointsHaveNotBeenSet;
	}

	/**
	 * Create a chain of transformations. One after the other.
	 * 
	 * @param transformations
	 *            the transformations to be chained
	 * @return the chain as graph
	 */
	public static TransformationGraph createChain(
			ObjectFactory... transformations) throws Exception {

		TransformationGraph result = new TransformationGraph();
		BoostedNode oldNode = null;
		for (ObjectFactory trafo : transformations) {

			BoostedNode newNode = result.addNode(trafo, true);

			if (oldNode != null) {
				Refactor.addEdge(oldNode, newNode);
			}

			oldNode = newNode;
		}


		return result;

	}

	@Override
	public void fillDocumentation(Object type, BoostedNode documentationParent,
			DocumentationContext context) throws Exception {
		if (DocumentationResourcesId.DOCUMENTATION_USER.equals(type)) {
			// TODO fill with general user description
			if (getRootNode() != null) {
				fillDocumentation(
						DocumentationResourcesId.DOCUMENTATION_USER_INSTANCE_DESCRIPTION,
						documentationParent, context);
			}
		} else if (DocumentationResourcesId.DOCUMENTATION_USER_INSTANCE_DESCRIPTION
				.equals(type)) {
			BoostedNode chapter = BookBoostUtils.createChapter(
					documentationParent, "Transformation Graph Description");
			BookBoostUtils
					.createText(
							chapter,
							"The graphic displays an overview of the graph, containing only the top-level-nodes.");
			// TODO activate that image later
			// BookBoostUtils.createImage(chapter,
			// JungGraphBoostUtils.createOverviewGraphComponent(this), "");//
			// "Overview over the graph"
			BookBoostUtils
					.createDiv(
							chapter,
							"The next graphic displays a detailed  view of the graph. Additionally to the top-level-nodes, all IN- and OUT-nodes belonging to the top-level-nodes are displayed.");
			// TODO activate that image later
			// BookBoostUtils.createImage(chapter,
			// JungGraphBoostUtils.createDefaultGraphComponent(this), "");//
			// "The graph with IN- and OUT- nodes"

			// BookBoostUtils.createDiv(chapter,
			// "TODO the content of this chapter");

		} else {
			throw ThrowableBoostUtils
					.createNotImplementedException("documentation generation");
		}
	}

	public BoostedNodeGraph getGraph() {
		return graph;
	}

	public void setGraph(BoostedNodeGraph graph) {
		this.graph = graph;
	}

	/**
	 * Create a default display configuration. The display may be used for
	 * debugging purposes.
	 * 
	 * @return the created configuration
	 */
	public static GraphDisplayConfiguration<TransformationNode<?>, String> createDefaultDisplayConfiguration() {
		GraphDisplayConfiguration<TransformationNode<?>, String> displayConfiguration = GraphDisplayConfiguration
				.createFilledNodesWithoutTextConfiguration();

		displayConfiguration
				.setVertexBackgroundTransformer(new Transformer<TransformationNode<?>, Paint>() {

					@Override
					public Paint transform(TransformationNode<?> input) {

						Object state = input.getComputationState();
						if (TransformationNode.STATE_OK.equals(state)) {
							return Color.GREEN;
						} else if (TransformationNode.STATE_NOT_YET_HANDLED
								.equals(state)) {
							return Color.BLUE;
						} else if (TransformationNode.STATE_ERROR.equals(state)) {
							return Color.RED;
						} else {
							return Color.BLACK;
						}

					}
				});

		displayConfiguration.setVertexDetailDisplayLevel(10);
		return displayConfiguration;
	}

	/**
	 * Create a simple visual display of this graph.
	 */
	public void simpleDisplay() {
		simpleDisplay(createDefaultDisplayConfiguration());
	}

	/**
	 * Create a simple visual display of this graph.
	 * 
	 * @param displayConfiguration
	 *            the configuration of the display
	 */
	public void simpleDisplay(GraphDisplayConfiguration displayConfiguration) {
		graph.simpleDisplay(displayConfiguration);
	}

	public boolean isDisplayTheGraphWhenAnErrorOccurs() {
		return displayTheGraphWhenAnErrorOccurs;
	}

	public void setDisplayTheGraphWhenAnErrorOccurs(
			boolean displayTheGraphWhenAnErrorOccurs) {
		this.displayTheGraphWhenAnErrorOccurs = displayTheGraphWhenAnErrorOccurs;
	}


	public boolean isContinuousTransformation() {
		return continuousTransformation;
	}

	public void setContinuousTransformation(boolean continuousTransformation) {
		this.continuousTransformation = continuousTransformation;
	}

	@Override
	public Object executeCommand(Object commandName, Object parameter) throws Exception {
		if (BasicCommands.COMMAND_START.equals(commandName)) {
			return getObjectInstance(parameter, null, null, null);
		} else {
			throw new IllegalStateException("command not supported: " + commandName);
		}

	}
}
