package net.sf.gluebooster.demos.pojo.refactor;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.naming.spi.ObjectFactory;

import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.basic.container.SimpleBoostedNode;
import net.sf.gluebooster.java.booster.essentials.meta.HasName;

/**
 * A node that contains a transformation.
 * 
 * @author CBauer
 * 
 * @defaultParamText inOrOutPoint the affected point
 * 
 */
public class TransformationNode<Name> extends SimpleBoostedNode<Name> {

	/**
	 * Attributename of the current values of in- and out-points of the nodes.
	 */
	private static final String CURRENT_VALUE = "currentValue";

	/**
	 * Attributename that the current values of in- and out-points of the nodes
	 * has been set (probably modified). The attribute is ignored when the
	 * computeTransformationsEvenIfInPointsHaveNotBeenSet is true.
	 */
	private static final String CURRENT_VALUE_SET = "currentValueSet";

	/**
	 * Attribute name. The value is the current computation state of this node.
	 */
	private static final String COMPUTATION_STATE = "computation state";

	/**
	 * The computation state that nothing has been done yet.
	 */
	public static final String STATE_NOT_YET_HANDLED = "not yet handled";

	/**
	 * The computation state that the computation has been done and everything
	 * is ok.
	 */
	public static final String STATE_OK = "ok";

	/**
	 * The computation state that the computation has been done but an error
	 * occurred.
	 */
	public static final String STATE_ERROR = "error";

	/**
	 * Are the out nodes to have the same result. If false, the transformation
	 * must have a map as result and the keys of the map must be the names of
	 * the out-points.
	 */
	private boolean sameResultInAllOutNodes = true;

	/**
	 * Sets the transformation of a node.
	 * 
	 * @param transformation
	 *            the transformation
	 */
	@SuppressWarnings("unchecked")
	public TransformationNode(Object transformation) {
		setTransformationAndName(transformation);

	}

	/**
	 * Get the transformation of one node.
	 * 
	 * @param node
	 *            contains the transformation
	 * @return the transformation is an ObjectFactory
	 */
	public ObjectFactory getTransformation() {
		return (ObjectFactory) getAttributes().getAttributes().get(
				ObjectFactory.class);
	}

	/**
	 * sets the transformation and the name
	 * 
	 * @param transformation
	 */
	public void setTransformationAndName(Object transformation) {
		TransformingDelegate delegate = new TransformingDelegate(transformation);
		getAttributes().getAttributes().put(ObjectFactory.class, delegate);

		getAttributes().setName(delegate.getName());


	}

	/**
	 * Sets the value of the attribute CURRENT_VALUE. Sets the CURRENT_VALUE_SET
	 * to true
	 * 
	 * @param value
	 *            the new value
	 */
	public static void setCurrentValue(BoostedNode inOrOutPoint, Object value) {
		inOrOutPoint.getAttributes().getAttributes().put(CURRENT_VALUE, value);
		setCurrentValueSet(inOrOutPoint, true);
	}

	public void setCurrentValue(Object value) {
		setCurrentValue(this, value);
	}
	/**
	 * Gets the value of the attribute CURRENT_VALUE.
	 * 
	 * @return the value of the point
	 */
	public static Object getCurrentValue(BoostedNode inOrOutPoint) {
		if (!hasCurrentValue(inOrOutPoint))
			throw new IllegalStateException("no current value yet defined");

		return inOrOutPoint.getAttributes().getAttributes().get(CURRENT_VALUE);
	}

	/**
	 * Gets the value of the attribute CURRENT_VALUE.
	 * 
	 * @return the value
	 */
	public Object getCurrentValue() {
		return getCurrentValue(this);
	}

	/**
	 * Has the point an attribute CURRENT_VALUE.
	 * 
	 * @return true if the attribute is set.
	 */
	public static boolean hasCurrentValue(BoostedNode inOrOutPoint) {
		return inOrOutPoint.getAttributes().getAttributes()
				.containsKey(CURRENT_VALUE);
	}

	/**
	 * Removes the current value of a given point
	 * 
	 * @param inOrOutPoint
	 *            the point which value should be removed
	 */
	public static void removeCurrentValue(BoostedNode inOrOutPoint) {
		inOrOutPoint.getAttributes().getAttributes().remove(CURRENT_VALUE);
		setCurrentValueSet(inOrOutPoint, false);
	}

	/**
	 * Sets the attribute CURRENT_VALUE_SET to true.
	 * 
	 * @param flag
	 *            the new value of the attribute
	 */
	public static void setCurrentValueSet(BoostedNode inOrOutPoint,
			boolean flag) {
		inOrOutPoint.getAttributes().getAttributes()
				.put(CURRENT_VALUE_SET, flag);

	}

	/**
	 * Gets the attribute CURRENT_VALUE_SET .
	 * 
	 * @return false if the attribute is not present or false.
	 */
	public static boolean isCurrentValueSet(BoostedNode inOrOutPoint) {
		Object result = inOrOutPoint.getAttributes().getAttributes()
				.get(CURRENT_VALUE_SET);
		if (result == null)
			return false;
		else
			return ((Boolean) result).booleanValue();

	}

	/**
	 * Returns the current values of the inpoints
	 * 
	 * @param returnMapEvenIfNotAllValuesAreKnown
	 *            should a map be returned even if not all values are known.
	 * @return null if not all values are known and
	 *         returnMapEvenIfNotAllValuesAreKnown == false; otherwise a map
	 *         with keys the names of the inpoint and as value their value
	 */
	public Map getInpointValues(boolean returnMapEvenIfNotAllValuesAreKnown) {

		List<BoostedNode> inPoints = getInPoints();

		HashMap inpointValueMap = new HashMap<Object, Object>();
		// the value to transform
		// it contains
		// of the inpoint

		boolean inpointValuesAreKnown = true;
		// initial value, it may be set to false later

		for (BoostedNode inPoint : inPoints) {
			// The inPoints must be instances of HasName

			boolean currentValueSet = isCurrentValueSet(inPoint);
			inpointValuesAreKnown &= currentValueSet;

			Object name = ((HasName) inPoint).getName();

			if (currentValueSet) {
				inpointValueMap.put(name, getCurrentValue(inPoint));
			} else if (returnMapEvenIfNotAllValuesAreKnown) {
				inpointValueMap.put(name, null);
			}

		}
		if (inpointValuesAreKnown || returnMapEvenIfNotAllValuesAreKnown)
			return inpointValueMap;
		else
			return null;
	}

	/**
	 * Removes the current value from all out-points.
	 */
	public void removeOutpointValues() {
		for (BoostedNode outPoint : getOutPoints()) {
			removeCurrentValue(outPoint);
		}
	}

	/**
	 * Removes the current value from all in-points
	 */
	public void removeInpointValues() {
		for (BoostedNode point : getInPoints()) {
			removeCurrentValue(point);
		}
	}

	/**
	 * Gets the computation state of this node.
	 * 
	 * @return the value of the attribute COMPUTATION_STATE
	 */
	public Object getComputationState() {
		return getAttributes().getFromAttributes(COMPUTATION_STATE);
	}

	/**
	 * Sets the computation state of this node.
	 * 
	 * @param value
	 *            the new value
	 */
	public void setComputationState(String value) {
		getAttributes().setAttribute(COMPUTATION_STATE, value);
	}

	/**
	 * Sets the computations state to STATE_NOT_YET_HANDLED.
	 */
	public void setStateNotYetHandled() {
		setComputationState(STATE_NOT_YET_HANDLED);
	}

	/**
	 * Is the computations state STATE_NOT_YET_HANDLED.
	 * 
	 * @return true if the state is STATE_NOT_YET_HANDLED
	 */
	public boolean isStateNotYetHandled() {
		return STATE_NOT_YET_HANDLED.equals(getComputationState());
	}

	/**
	 * Sets the computations state to STATE_OK.
	 */
	public void setStateOk() {
		setComputationState(STATE_OK);
	}

	/**
	 * Sets the computations state to STATE_ERROR.
	 */
	public void setStateError() {
		setComputationState(STATE_ERROR);
	}

	/**
	 * Uses the in-point values to do a transformation and sets the outpoint
	 * values
	 * 
	 * @param computeTransformationsEvenIfInPointsHaveNotBeenSet
	 *            should a result be computed even if not all information of the
	 *            points is present (maybe a default value is used in that case)
	 * @return true, if at least one new value has been computed.
	 */
	public boolean transform(boolean computeTransformationsEvenIfInPointsHaveNotBeenSet) throws Exception
	{
		try 
		{
			Map inpointValueMap = getInpointValues(computeTransformationsEvenIfInPointsHaveNotBeenSet);

			if (inpointValueMap == null) {
				return false;
			} else {
			
				Object value;
				// the value to transform
				// there are special cases:
				switch (inpointValueMap.size()) {
				case 0:
					value = null;
					break;
				case 1:
					value = inpointValueMap.values().iterator().next();
					// value is directly the content of the map
					break;
				default:
					value = inpointValueMap;
				}
			
				Object newValue = getTransformation()
						.getObjectInstance(value, null, null, null);
			
				// set the values of the node and its out points
				setCurrentValue(newValue);
			
				List<BoostedNode> outPoints = getOutPoints();
				switch (outPoints.size()) {
				case 0: /* nothing to do */
					break;
				case 1:
					TransformationNode.setCurrentValue(
							outPoints.get(0), newValue);
					break;
				default:
					for (BoostedNode outPoint : outPoints) {
						if (sameResultInAllOutNodes) {
							TransformationNode.setCurrentValue(outPoint,
									newValue);
						} else {
							Map map = (Map) newValue;
							TransformationNode.setCurrentValue(outPoint,
									map.get(((HasName) outPoint)
												.getName()));
						}
					}
				}
				
				setStateOk();
				return true;
			}
		
		} catch (Exception ex) {
			setStateError();
			throw ex;
		}
	}

	public boolean isSameResultInAllOutNodes() {
		return sameResultInAllOutNodes;
	}

	public void setSameResultInAllOutNodes(boolean sameResultInAllOutNodes) {
		this.sameResultInAllOutNodes = sameResultInAllOutNodes;
	}

}
