package net.sf.gluebooster.demos.pojo.refactor;

import javax.naming.spi.ObjectFactory;

import net.sf.gluebooster.java.booster.essentials.utils.ReflectionBoostUtils;

/**
 * Transforms object by invoking a method, executing an ObjectFactory without
 * arguments, ...
 * 
 * @author CBauer
 *
 */
public class TransformerByInvoking<Source, Target> extends
		BoostedObjectFactory<Source, Target> {

	/**
	 * Read all of a stream, etc.
	 */
	private static final int TYPE_INVOKE_OBJECT_FACTORY_WITHOUT_ARGUMENTS = 1;

	/**
	 * Invoke a configured action method on the stored object.
	 */
	private static final int TYPE_INVOKE_GIVEN_METHOD_ON_GIVEN_OBJECT = 2;

	/**
	 * Invoke a new instance of the stored object (must be a class).
	 */
	private static final int TYPE_INVOKE_NEW_INSTANCE = 3;

	/**
	 * The type of the transformer
	 */
	private int type;

	/**
	 * A stored object.
	 */
	private Object object;

	/**
	 * The action (method) to do.
	 */
	private String action;

	/**
	 * Creates a transformer that needs an ObjectFactory as argument which it invokes.
	 * 
	 * @return the created transformer
	 */
	public static <Source, Target> BoostedObjectFactory<Source, Target> createInvokingObjectFactoryWithoutArgumentsTransformer() {
		TransformerByInvoking result = new TransformerByInvoking(
				TYPE_INVOKE_OBJECT_FACTORY_WITHOUT_ARGUMENTS);
		return result;
	}

	/**
	 * Creates a transformer that invokes a method on an object. The parameter of the method is the argument of the transformation method.
	 *
	 * @param object
	 *            if the object is not null then the method is invoked with the parameter supplied by transformWithException. if the object is null, then the
	 *            parameter supplied by transformWithException is used as object
	 * @param methodname
	 *            the name of the method
	 * @return the created transformer
	 */
	public static <Source, Target> BoostedObjectFactory<Source, Target> createMethodInvokingTransformer(Object object, String methodname) {
		TransformerByInvoking result = new TransformerByInvoking(TYPE_INVOKE_GIVEN_METHOD_ON_GIVEN_OBJECT);
		result.object = object;
		result.action = methodname;
		return result;
	}

	/**
	 * Creates a transformer that creates new instances of a given class
	 * 
	 * @param instanceClass
	 *            the class to generate instances
	 * @return the created transformer
	 */
	public static <Source, Target> BoostedObjectFactory<Source, Target> createNewInstanceInvokingTransformer(Class instanceClass) {
		TransformerByInvoking result = new TransformerByInvoking(TYPE_INVOKE_NEW_INSTANCE);
		result.object = instanceClass;
		return result;
	}

	private TransformerByInvoking(int type) {
		this.type = type;
	}

	@Override
	protected Target transformWithException(Source input) throws Exception {

		switch (type) {
		case TYPE_INVOKE_OBJECT_FACTORY_WITHOUT_ARGUMENTS:
			return (Target) ((ObjectFactory) input).getObjectInstance(null,
					null, null, null);
		case TYPE_INVOKE_GIVEN_METHOD_ON_GIVEN_OBJECT:
			Object result;
			if (object != null) {
				result = ReflectionBoostUtils.getMethod(object.getClass(), action).invoke(object, input);
			} else {
				result = ReflectionBoostUtils.getMethod(input.getClass(), action).invoke(input);
			}
			// System.out.println("TYPE_INVOKE_GIVEN_METHOD_ON_GIVEN_OBJECT needs to be implemented");
			return (Target) result;
		case TYPE_INVOKE_NEW_INSTANCE:
			return (Target) ((Class) object).newInstance();
		default:
			throw new IllegalStateException("Type not supported: " + type);

		}
	}

	public Object getObject() {
		return object;
	}

	public void setObject(Object object) {
		this.object = object;
	}

	public String getAction() {
		return action;
	}

	public void setAction(String action) {
		this.action = action;
	}

}
