package net.sf.gluebooster.demos.pojo.refactor;

import java.util.Map;

import org.apache.commons.collections15.Transformer;

/**
 * Transforms by getting/putting values from/to a map.
 * 
 * @author cbauer
 *
 * @param <Key>
 * @param <Value>
 */
public class TransformerByMap<Key, Value> implements Transformer<Key, Value>, Runnable {

	/**
	 * The type to get a value from the map. Key is the input of the transformation
	 */
	private static final int TYPE_GET_FROM_MAP = 1;

	private static final int TYPE_PUT_CONSTANT_INTO_MAP = 2;

	/**
	 * The type of the transformation
	 */
	private int type;

	/**
	 * The map with all key/values.
	 */
	private Map<Key, Value> map;

	/**
	 * A key of the map used for the transformation
	 */
	private Key key;

	/**
	 * A value used for the transformation
	 */
	private Value value;

	private TransformerByMap() {

	}


	/**
	 * Creates a transformer of type TYPE_GET_FROM_MAP.
	 * 
	 * @return the created transformer
	 */
	public static <Key, Value> TransformerByMap<Key, Value> createGetFromMapTransformer(Map<Key, Value> map) {
		TransformerByMap<Key, Value> result = new TransformerByMap<>();
		result.type = TYPE_GET_FROM_MAP;
		result.map = map;
		return result;
	}

	/**
	 * Creates a transformer that puts a constant into a map
	 * 
	 * @param map
	 *            the map into the constant is put
	 * @param key
	 *            the key of the map
	 * @param value
	 *            the new value
	 * @return the created transformer
	 */
	public static <Key, Value> TransformerByMap<Key, Value> createPutConstantIntoMapTransformer(Map<Key, Value> map, Key key, Value value) {
		TransformerByMap<Key, Value> result = new TransformerByMap<>();
		result.type = TYPE_PUT_CONSTANT_INTO_MAP;
		result.map = map;
		result.key = key;
		result.value = value;
		return result;
	}

	@Override
	public Value transform(Key input) {

		if (type == TYPE_GET_FROM_MAP) {
			return map.get(input);
		} else if (type == TYPE_PUT_CONSTANT_INTO_MAP) {
			map.put(key, value);
			return null;
		} else {
			throw new IllegalStateException("type not supported: " + type);
		}
	}

	@Override
	public void run() {
		transform(null);

	}

}
