package net.sf.gluebooster.demos.pojo.planning;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.collections15.Transformer;
import org.junit.Assert;
import org.junit.Test;

import net.sf.gluebooster.demos.pojo.planning.blockworld.Block;
import net.sf.gluebooster.demos.pojo.planning.blockworld.Blocktype;
import net.sf.gluebooster.demos.pojo.planning.blockworld.Blockworld;
import net.sf.gluebooster.demos.pojo.planning.blockworld.Hand;
import net.sf.gluebooster.demos.pojo.planning.blockworld.Orientation;
import net.sf.gluebooster.demos.pojo.planning.blockworld.Table;
import net.sf.gluebooster.demos.pojo.planning.blockworld.TableElement;
import net.sf.gluebooster.demos.pojo.refactor.BoostedHashSetExtension;
import net.sf.gluebooster.demos.pojo.refactor.ExtendedTestRoot;
import net.sf.gluebooster.demos.pojo.refactor.GraphDisplayConfiguration;
import net.sf.gluebooster.demos.pojo.refactor.Operation;
import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.basic.gui.swing.JOptionPaneUserInteraction;
import net.sf.gluebooster.java.booster.essentials.logging.LogBooster;
import net.sf.gluebooster.java.booster.essentials.meta.Example;
import net.sf.gluebooster.java.booster.essentials.objects.BoostedObject;

public class PlanningTest extends ExtendedTestRoot implements ActionListener {


	private String getDefaultBlockname(int blocknumber){

		return  "Block " + blocknumber;
	}
	
	
	private Block getDefaultBlock(int blocknumber, int x, int y,
			Orientation orientation, boolean up, int[] overBlockNumbers) {
		String blockname = getDefaultBlockname(blocknumber);
		Blocktype type;
		switch (blocknumber) {
		case 1:
			type = Blocktype.QUADRUPLEBLOCK;
			break;
		case 2:
			type = Blocktype.DOUBLEBLOCK;
			break;
		default:
			type = Blocktype.SINGLEBLOCK;
			break;
		}
		HashSet<String> overBlocks = new HashSet<String>();
		if (overBlockNumbers != null) {
			for (int overBlock : overBlockNumbers) {
				overBlocks.add(getDefaultBlockname(overBlock));
			}
		}

		return new Block(blockname, x, y, type, orientation, up, overBlocks);

	}
	
	private Hand getDefaultHand() {
		Hand result = new Hand("Hand 1", 0, 0);
		result.setUp(Boolean.TRUE);
		return result;
	}

	private Table getDefaultTable(Set<TableElement> blocksAndHands) {
		int width = 7;
		int depth = 5;
		return new Table(width, depth, /* agents, */blocksAndHands);

	}

	/**
	 * Creates a pyramid of blocks
	 */
	private Planning<Table, String> createPyramidTestdata(int numberOfBlocks,
			int stepsToRun)
			throws Exception {


		Set<TableElement> initialBlocksAndHands = new HashSet<TableElement>();
		Set<TableElement> wantedBlocksAndHands = new HashSet<TableElement>();

		initialBlocksAndHands.add(getDefaultHand());

		for (int i = 1; i <= numberOfBlocks; i++) {

			int[] overBlocks = null;
			if (i > 1) {
				overBlocks = new int[] { (i - 1) };
			}

			initialBlocksAndHands.add(getDefaultBlock(i, 0, 0, Orientation.Y,
					false, overBlocks));
			wantedBlocksAndHands.add(getDefaultBlock(i, 2, 0, Orientation.Y,
					false, overBlocks));

		}

		Plan<Table, String> plan = new Plan<Table, String>(new Blockworld(),
				getDefaultTable(initialBlocksAndHands),
				getDefaultTable(wantedBlocksAndHands));

		PlannerEnv env = new PlannerEnv();
		env.setPlanningLog(new LogBooster("planner"));
		env.setMaxSteps(stepsToRun);


		return new Planning<Table, String>(plan, env);
	}

	@Test
	public void testIdlePlanner() throws Exception {

		Planning<Table, String> data = createPyramidTestdata(3, 100);

		// ObjectFactory has the signature Planning getObjectInstance (Planning,
		// ...)
		Transformer<Planning<Table, String>, ?> planner = MetaPlanner
				.createIdlePlanner();
		planner.transform(data);

	}

	/**
	 * Tests that the planner recognizes a situation that is already solved.
	 * 
	 * @throws Exception
	 */
	@Test
	public void testNoNeedToDoAnything() throws Exception {
		Planning<Table, String> data = createPyramidTestdata(3, 1);
		// make the postcondition equal to the precondition (only without the hand)
		Operation<Table,String> riddle = data.getPlan().getOperation(
				data.getPlan().getUnsolvedNodes().iterator().next());
		Table condition = riddle.getGlobalPrecondition().cloneMe();
		Hand hand = condition.getHands().iterator().next();
		condition.removeTableElement(hand);
		riddle.setPostcondition(condition);
		
		Transformer<Planning<Table, String>, Planning<Table, String>> planner = MetaPlanner.createPlanner1();		
		
		for (int i = 0; i <100; i++) {
			data = planner.transform(data);// do some steps
		}
		
		data.getPlan().simpleDisplay();

		Assert.assertTrue("The plan should be solved.", data.getPlan().isSolved());
		Assert.assertEquals(
				"There should be only one node in the plan, because nothing is to do",
				1, data.getPlan().getGraph().getAllVertices().size());
		
	}

	@Test
	public void testPlanner1() throws Exception {
		testPlanner(MetaPlanner.createPlanner1(), 5, 30);
	}

	private Planning<Table, String> createPyramidRiddle(int numberOfBlocks, StringBuilder operationsLog, StringBuilder resultLog) throws Exception {

		Planning<Table, String> data = createPyramidTestdata(numberOfBlocks, 1);
		data.setName("Pyramid-Riddle with " + numberOfBlocks + " blocks");

		JOptionPaneUserInteraction interaction = new JOptionPaneUserInteraction();

		GraphDisplayConfiguration<BoostedNode, String> displayConfiguration = data.getPlan().getSolveStateDisplayConfiguration();
		data.setDefaultDisplayConfiguration(displayConfiguration);
		data.defaultDisplay();
		BoostedHashSetExtension<BoostedNode> nodes = data.getPlan().getGraph().getSomeNodesOfTheConnectedComponents();
		nodes.setInformListenerWhenEmpty(true);
		nodes.getListeners().addActionListener(this);
		BoostedNode riddle = nodes.iterator().next();

		String riddleString = "Riddle\n" + data.getPlan().getOperation(riddle).toString(3) + "\n=============\n\n";

		if (operationsLog != null) {
			operationsLog.append(riddleString);
		}

		if (resultLog != null) {
			resultLog.append(riddleString);
		}

		return data;
	}

	@Test
	@Example(clasz = Blockworld.class)
	public void pyramidDemo() throws Exception {

		StringBuilder resultLog = new StringBuilder();
		// [gb:useInDocumentation]
		// An example how the planner relocates a pyramid.
		Planning<Table, String> data = createPyramidRiddle(3, null, resultLog);
		solve(data, MetaPlanner.createPlanner1(), 1000, 1000, null, resultLog);
		// [/gb:useInDocumentation]

		data.getPlan().logOperatorsAndDetails(resultLog);
		writeExampleResult(resultLog);


	}

	private void solve(Planning<Table, String> data, Transformer thePlanner, int firstStepsWithoutDisplay, int stepsWithDisplay, StringBuilder operationsLog,
			StringBuilder resultLog) throws Exception {

		Transformer<Planning<Table, String>, Planning<Table, String>> planner = (Transformer<Planning<Table, String>, Planning<Table, String>>) thePlanner;

		boolean solved = false;
		int stepCounter = 0;

		ArrayList<String> operatorTransformation = new ArrayList();

		StringBuilder operators;

		// do {
		try {
			for (int i = firstStepsWithoutDisplay; i > 0; i--) {
				stepCounter++;
				data = planner.transform(data);// do some steps
			}

			for (int i = stepsWithDisplay; i > 0 && !solved; i--) {
				stepCounter++;
				data.defaultDisplay();
				data = planner.transform(data);// do some steps
				solved = data.getPlan().isSolved();
				if (operationsLog != null) {
					operationsLog.append("Step " + stepCounter + "\n\n");
				}
				operators = new StringBuilder();
				data.getPlan().logOperatorsAndDetails(operators);
				operatorTransformation.add(operators.toString());
				if (operationsLog != null) {
					operationsLog.append(operators);
					operationsLog.append("\n-----------------\n\n");
				}
			}
		} finally {
			data.defaultDisplay();
		}

		if (operationsLog != null) {
			getLog().info(operationsLog);
		}
		// } while ("".equals(interaction.showInputDialog("Next Step",
		// null)));
		// Thread.sleep(10000);

		Assert.assertTrue("The result plan should be solved. " + data.getName(), solved);

	}

	private void testPlanner(Transformer thePlanner,
			int firstStepsWithoutDisplay, int stepsWithDisplay)
			throws Exception {

		int minNumberOfBlocks = 3;
		int maxNumberOfBlocks = 3;


		for (int numberOfBlocks = minNumberOfBlocks; numberOfBlocks <= maxNumberOfBlocks; numberOfBlocks++) {

			StringBuilder operations = new StringBuilder();
			StringBuilder result = new StringBuilder();

			Planning<Table, String> data = createPyramidRiddle(numberOfBlocks, operations, result);

			solve(data, thePlanner, firstStepsWithoutDisplay, stepsWithDisplay, operations, result);

		}
	}

	@Test
	public void testPlannerThatJustExpands() throws Exception {
		testPlanner(MetaPlanner.createPlannerThatJustExpands(), 18, 40);
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		getLog().info("actionPerformed", e);

	}

	@Test
	public void testTableGuiString() throws Exception {
		Planning<Table, String> planning = createPyramidTestdata(3, 1);
		Operation<Table, String> operation = planning.getPlan().getOperation(
				planning.getPlan().getGraph().getAllVertices().iterator()
						.next());

		String gui = operation.getGlobalPrecondition().getGuiString();
		Assert.assertNotNull(gui);
		getLog().info(gui);
	}

}
