package net.sf.gluebooster.demos.pojo.refactor;

import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.Line2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import junit.framework.Assert;
import net.sf.gluebooster.java.booster.basic.container.BoostedNode;
import net.sf.gluebooster.java.booster.basic.container.SimpleBoostedNode;

import org.apache.commons.collections15.Transformer;
import org.apache.commons.lang3.tuple.Pair;
import org.junit.Test;

import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.UndirectedSparseGraph;

public class GeometryBoostUtilsTest {


	@Test
	public void testRepositionWithoutOverlapping() {
		Rectangle r1 = new Rectangle(0, 0, 5, 2);
		Rectangle r2 = new Rectangle(4, 0, 4, 3);
		Rectangle r3 = new Rectangle(1, 1, 5, 3);
		
		
		//  y
		//  ^    3   3   3   3   3   3
		//  |                                             
		//  |                                              
		//  |                        3                     
		// 3|    3            2  2   2   2   2          
		//  |                                           
		//  |                                           
		//  |    3                                       
		// 2|1   1   1   1   12  1   3       2          
		//  |                                            
		//  |                                            
		//  |                                             
		// 1|1   3   3   3   32  13  3       2          
		//  |                                          
		//  |                                             
		//  |1   1   1   1   12  12  2   2   2       
		//---------------------------------------------->  x
		//  0    1   2   3   4   5   6   7   8   9   10
		
		Graph<Shape, String> graph = new UndirectedSparseGraph<Shape, String>();
		graph.addVertex(r1);
		graph.addVertex(r2);
		graph.addVertex(r3);
		graph.addEdge("12", r1, r2);
		graph.addEdge("23", r2, r3);
		graph.addEdge("31", r3, r1);

		GeometryPositionConstraint<Object> constraints = new GeometryPositionConstraint<Object>();
		constraints.setMinimumDistance(2);
		GeometryBoostUtils.repositionWithoutOverlapping(graph, constraints);

		Assert.assertFalse(r1.intersects(r2));
		Assert.assertFalse(r1.intersects(r3));
		Assert.assertFalse(r2.intersects(r3));

		Line2D.Float line12 = new Line2D.Float(r1.x, r1.y, r2.x, r2.y);
		Line2D.Float line13 = new Line2D.Float(r1.x, r1.y, r3.x, r3.y);
		Line2D.Float line32 = new Line2D.Float(r3.x, r3.y, r2.x, r2.y);

		Assert.assertFalse(line12.intersects(r3));
		Assert.assertFalse(line13.intersects(r2));
		Assert.assertFalse(line32.intersects(r1));

	}

	private BoostedNodeGraph createSampleGraph(int numberOfNodes)
			throws Exception {

		int numberOfNodesMinus1 = numberOfNodes - 1;
		BoostedNodeGraph graph = new BoostedNodeGraph();

		Map<Integer, BoostedNode> nodeMapping = new HashMap<Integer, BoostedNode>();
		for (int i = -1; i < numberOfNodesMinus1; i++) {
			SimpleBoostedNode<Integer> node = new SimpleBoostedNode<Integer>();
			node.setName(i);
			nodeMapping.put(i, node);
			graph.addNode(node);
		}

		// some hardcoded edges for an interesting graph
		int[][] edges = { { 0, 1 }, { 0, 2 }, { 1, 3 }, { 3, 4 }, { 4, 5 },
				{ 2, 6 }, { 6, 7 }, { 7, 8 }, { 5, 9 }, { 9, 8 }, { 8, 10 } };

		for (int[] edge : edges) {
			if (edge[0] < numberOfNodesMinus1 && edge[1] < numberOfNodesMinus1) {
				graph.addEdge(nodeMapping.get(edge[0]),
						nodeMapping.get(edge[1]));
			}
		}
		for (int i = 11; i < numberOfNodesMinus1; i++) {
			graph.addEdge(nodeMapping.get(i - 1), nodeMapping.get(i));
		}

		return graph;
	}

	@Test
	public void testRepositionWithoutOverlapping2() throws Exception {

		int startnumber = 1;
		// int startnumber = 11;
		int maxnumber = 11;

		for (int numberOfNodes = startnumber; numberOfNodes < maxnumber; numberOfNodes++) {
			BoostedNodeGraph graph = createSampleGraph(numberOfNodes);

			GraphDisplayConfiguration<BoostedNode, String> configuration = GraphDisplayConfiguration
					.createUnadornedConfiguration();
			Transformer<BoostedNode, Shape> vertexShapeTransformer = new Transformer<BoostedNode, Shape>() {

				HashMap<BoostedNode, Rectangle> nodeMapping = new HashMap<BoostedNode, Rectangle>();

				@SuppressWarnings("unchecked")
				@Override
				public Shape transform(BoostedNode input) {

					int id = 0;
					if (input instanceof SimpleBoostedNode) {
						id = ((SimpleBoostedNode<Integer>) input).getName();
					}
					if (!nodeMapping.containsKey(input)) {
						nodeMapping.put(input, new Rectangle(100 + id, 100));
					}
					return nodeMapping.get(input);
				}

			};

			configuration.setVertexShapeTransformer(vertexShapeTransformer);
			configuration.setDisplayVertexLabel(true);
			configuration.setTitle("Graph " + numberOfNodes);
			graph.simpleDisplay(configuration, new Flag(false, "numberOfNodes="
					+ numberOfNodes + ": "));

		}
		int lineForDebugging = 0;
	}

	@Test
	public void testRepositionSpecialStructures() {

		int minVertexCount = 1;
		int maxVertexCount = 6;

		GeometryPositionConstraint<Shape> constraints = new GeometryPositionConstraint<Shape>();
		constraints.setMinimumDistance(50);

		for (int vertexCount = minVertexCount; vertexCount < maxVertexCount; vertexCount++) {

			// part 1 create a graph where all vertices are connected
			Graph<Shape, String> graph = new UndirectedSparseGraph<Shape, String>();

			ArrayList<Rectangle> vertices = new ArrayList<Rectangle>(
					vertexCount);

			for (int i = 0; i < vertexCount; i++) {
				Rectangle r = new Rectangle(0, 0, 50 + i * 10 + i, 50 + i * 10
						+ i);
				vertices.add(r);
				graph.addVertex(r);
			}

			// connect all vertices
			for (int i = 0; i < vertexCount; i++) {
				for (int j = i + 1; j < vertexCount; j++) {
					graph.addEdge("" + i + "-" + j, vertices.get(i),
							vertices.get(j));
				}
			}

			// part 2: reposition and check
			GeometryBoostUtils.repositionSpecialStructures(graph, constraints);

			Refactor.displayShapes(new DisplayConfiguration("Vertices: "
					+ vertexCount, graph));


			Triple<Collection<Pair<Shape, Shape>>, Collection<Triple<Shape, String, edu.uci.ics.jung.graph.util.Pair<Shape>>>, Collection<Pair<String, String>>> overlappings = GeometryBoostUtils
					.hasOverlappings(graph, false);
			Assert.assertTrue("overlapping vertices " + vertexCount,
					overlappings.getFirst().isEmpty());
			Assert.assertTrue("overlapping vertex with edge " + vertexCount,
					overlappings.getSecond().isEmpty());
			// edges may overlap
			// Assert.assertTrue("overlapping edges " + vertexCount,
			// overlappings
			// .getThird().isEmpty());

			// part 3
			// check with constraints
			Rectangle dimension = GeometryBoostUtils.getDimension(graph
					.getVertices());

			// create a pentagram with additional rectangles at the following positions of the dimension:
			// top left, top right, middle left, middle right, bottom center
			// and connect them (virtually) with some of the graph vertices and do a new layout with these additional constraints
			int size = 10;// the size of the new rectangles
			Rectangle[] outsideVertices = {
					new BoostedRectangle("top left", 0, 0, size, size),
					new BoostedRectangle("bottom center", dimension.width / 2,
							dimension.height + 3 * size, size, size),
					new BoostedRectangle("top right", dimension.width - size,
							0, size, size),
					new BoostedRectangle("left", 0, dimension.height / 2, size,
							size),
					new BoostedRectangle("right", dimension.width + 3 * size,
							dimension.height / 2, size, size), };

			ArrayList<Shape> edgesToLowerX = new ArrayList<Shape>();
			ArrayList<Shape> edgesToHigherX = new ArrayList<Shape>();
			ArrayList<Shape> edgesToLowerY = new ArrayList<Shape>();
			ArrayList<Shape> edgesToHigherY = new ArrayList<Shape>();

			if (vertexCount > 0) {
				edgesToLowerY.add(vertices.get(0));
			}
			if (vertexCount > 1) {
				edgesToHigherY.add(vertices.get(1));
			}
			if (vertexCount > 2) {
				edgesToLowerY.add(vertices.get(2));
			}
			if (vertexCount > 3) {
				edgesToLowerX.add(vertices.get(3));
			}
			if (vertexCount > 4) {
				edgesToHigherX.add(vertices.get(4));
			}

			Rectangle newBounds = new Rectangle(size + 10, size + 10,
					dimension.width, dimension.height);// just a translation of the size of the new rectangles
			GeometryPositionConstraint<Shape> constraint = new GeometryPositionConstraint<Shape>(
					edgesToLowerX, edgesToHigherX, edgesToLowerY,
					edgesToHigherY, newBounds);
			constraint.setMinimumDistance(constraints.getMinimumDistance());

			GeometryBoostUtils.repositionSpecialStructures(graph, constraint);

			// part 4
			// display a new graph with the additional outside vertices
			// the graph can not be used, because of the changed positions the hashcode of the graph vertices has changed and therefore the vertices are added
			// again
			
			Graph<Shape, String> graph2 = JungGraphBoostUtils.copy(graph);

			// new UndirectedSparseGraph<Shape, String>();
			// for (Shape vertex: graph.getVertices()){
			// graph2.addVertex(vertex);
			// }
			// for (String edge: graph.getEdges()){
			// edu.uci.ics.jung.graph.util.Pair<Shape> endpoints = graph
			// .getEndpoints(edge);
			// graph2.addEdge(edge, endpoints.getFirst(),
			// endpoints.getSecond());
			// }

			for (int i = 0; i < vertexCount; i++) {
				graph2.addVertex(outsideVertices[i]);
				graph2.addEdge("outside" + i, vertices.get(i),
						outsideVertices[i]);
			}

			Refactor.displayShapes(new DisplayConfiguration("new Vertices: ",
					graph2));

			overlappings = GeometryBoostUtils.hasOverlappings(graph2, false);

			// TODO activate these tests again
			if (false) {
				Assert.assertTrue("Overlapping (Vertex count " + vertexCount + "): " + overlappings.getFirst(), overlappings.getFirst().isEmpty());
				Assert.assertTrue("Overlapping (Vertex count " + vertexCount + "): " + overlappings.getSecond(), overlappings.getSecond().isEmpty());
			}

			// edges may overlap
			// Assert.assertTrue("Overlapping: " + overlappings.getThird(),
			// overlappings.getThird().isEmpty());

		}
	}

	@Test
	public void testPositionByDivideAndConquerWith3() throws Exception {
		Rectangle rect1 = new Rectangle(0, 0, 10, 10);
		Rectangle rect2 = new Rectangle(0, 0, 20, 20);
		Rectangle rect3 = new Rectangle(0, 0, 30, 30);
		Graph<Shape, String> graph = new UndirectedSparseGraph<Shape, String>();
		graph.addVertex(rect1);
		graph.addVertex(rect2);
		graph.addVertex(rect3);
		graph.addEdge("1-2", rect1, rect2);

		testPositionByDivideAndConquer(graph);
	}

	@Test
	public void testPositionByDivideAndConquerWith4() throws Exception {
		Rectangle rect1 = new BoostedRectangle("rect1", 0, 0, 10, 10);
		Rectangle rect2 = new BoostedRectangle("rect2", 0, 0, 20, 20);
		Rectangle rect3 = new BoostedRectangle("rect3", 0, 0, 30, 30);
		Rectangle rect4 = new BoostedRectangle("rect4", 0, 0, 40, 40);

		Graph<Shape, String> graph = new UndirectedSparseGraph<Shape, String>();
		graph.addVertex(rect1);
		graph.addVertex(rect2);
		graph.addVertex(rect3);
		graph.addVertex(rect4);
		graph.addEdge("1-2", rect1, rect2);
		graph.addEdge("1-3", rect1, rect3);


		testPositionByDivideAndConquer(graph);
	}

	private void testPositionByDivideAndConquer(Graph<Shape, String> graph) {
		GeometryPositionConstraint<Shape> constraints = new GeometryPositionConstraint<Shape>();
		constraints.setMinimumDistance(5);
		GeometryBoostUtils.positionByDivideAndConquer(graph, constraints);

		Refactor.displayShapes(new DisplayConfiguration("Divide and Conquer ",
				graph));

		Triple<Collection<Pair<Shape, Shape>>, Collection<Triple<Shape, String, edu.uci.ics.jung.graph.util.Pair<Shape>>>, Collection<Pair<String, String>>> overlapping = GeometryBoostUtils
				.hasOverlappings(graph, true);

		Assert.assertTrue(overlapping.getFirst().isEmpty());
		Assert.assertTrue(overlapping.getSecond().isEmpty());
		Assert.assertTrue(overlapping.getThird().isEmpty());

	}

	/**
	 * A line hat problems with returning it's correct x and y. The example is
	 * tested here.
	 */
	@Test
	public void testLineValues() {
		double a1 = 5.0;
		double b1 = 5.0;
		double a2 = 20.0;
		double b2 = 5.0;
		Line2D.Double line = new Line2D.Double(a1, b1, a2, b2);
		double x1 = line.getX1();
		double x2 = line.getX2();
		double y1 = line.getY1();
		double y2 = line.getY2();
		Assert.assertEquals(a1, x1, 0.01);
		Assert.assertEquals(a2, x2, 0.01);
		Assert.assertEquals(b1, y1, 0.01);
		Assert.assertEquals(b2, y2, 0.01);

		Rectangle rect = new Rectangle(55, 5, 20, 20);
		Assert.assertFalse(rect.intersectsLine(line));

	}
}
